@@ -87,8 +87,8 @@ use Time::HiRes qw(time);
 use Cwd;
 use Config;
 
-our $VERSION = "4.000000";      # update after release (same format as perl $])
-#our $IS_DEVEL_BUILD = 1;        # 1 for devel build 
+our $VERSION = "4.000002";      # update after release (same format as perl $])
+# our $IS_DEVEL_BUILD = 1;        # 1 for devel build 
 our $IS_DEVEL_BUILD = 0;        # 0 for release versions including rc & pre releases
 
 # Used during the prerelease/release-candidate part of the official release
@@ -100,18 +100,18 @@ our @ISA = qw();
 
 # SUB_VERSION is now just <yyyy>-<mm>-<dd>
 our $SUB_VERSION = 'svnunknown';
-if ('$LastChangedDate: 2022-12-14 02:29:30 +0000 (Wed, 14 Dec 2022) $' =~ ':') {
-  # Subversion keyword "$LastChangedDate: 2022-12-14 02:29:30 +0000 (Wed, 14 Dec 2022) $" has been successfully expanded.
+if ('$LastChangedDate: 2025-08-27 15:15:53 +0200 (Wed, 27 Aug 2025) $' =~ ':') {
+  # Subversion keyword "$LastChangedDate: 2025-08-27 15:15:53 +0200 (Wed, 27 Aug 2025) $" has been successfully expanded.
   # Doesn't happen with automated launchpad builds:
   # https://bugs.launchpad.net/launchpad/+bug/780916
-  $SUB_VERSION = (split(/\s+/,'$LastChangedDate: 2022-12-14 02:29:30 +0000 (Wed, 14 Dec 2022) $ updated by SVN'))[1];
+  $SUB_VERSION = (split(/\s+/,'$LastChangedDate: 2025-08-27 15:15:53 +0200 (Wed, 27 Aug 2025) $ updated by SVN'))[1];
 }
 
 
 if (defined $IS_DEVEL_BUILD && $IS_DEVEL_BUILD) {
-  if ('$LastChangedRevision: 1905971 $' =~ ':') {
-    # Subversion keyword "$LastChangedRevision: 1905971 $" has been successfully expanded.
-    push(@EXTRA_VERSION, ('r' . qw{$LastChangedRevision: 1905971 $ updated by SVN}[1]));
+  if ('$LastChangedRevision: 1928046 $' =~ ':') {
+    # Subversion keyword "$LastChangedRevision: 1928046 $" has been successfully expanded.
+    push(@EXTRA_VERSION, ('r' . qw{$LastChangedRevision: 1928046 $ updated by SVN}[1]));
   } else {
     push(@EXTRA_VERSION, ('r' . 'svnunknown'));
   }
@@ -173,7 +173,7 @@ our @default_userstate_dir = (
 
 ###########################################################################
 
-=item $t = Mail::SpamAssassin->new( { opt => val, ... } )
+=item $t = Mail::SpamAssassin-E<gt>new( { opt =E<gt> val, ... } )
 
 Constructs a new C<Mail::SpamAssassin> object.  You may pass a hash
 reference to the constructor which may contain the following attribute-
@@ -272,7 +272,7 @@ in advance that some information will not be needed by a caller.
 A value of the option can either be a string (a comma-delimited list of tag
 names), or a reference to a list of individual tag names. A caller may provide
 the list in advance, specifying his intention to later collect the information
-through $pms->get_tag() calls. If a name of a tag starts with a 'NO' (case
+through $pms-E<gt>get_tag() calls. If a name of a tag starts with a 'NO' (case
 insensitive), it shows that a caller will not be interested in such tag,
 although there is no guarantee it would save any resources, nor that a tag
 value will be empty. Currently no built-in tags start with 'NO'. A later
@@ -289,9 +289,9 @@ Not requesting it can save a millisecond or two - it mostly serves to
 illustrate the usage of need_tags.
 
 Example:
-  need_tags =>    'TIMING,noLANGUAGES,RELAYCOUNTRY,ASN,noASNCIDR',
+  need_tags =E<gt>    'TIMING,noLANGUAGES,RELAYCOUNTRY,ASN,noASNCIDR',
 or:
-  need_tags => [qw(TIMING noLANGUAGES RELAYCOUNTRY ASN noASNCIDR)],
+  need_tags =E<gt> [qw(TIMING noLANGUAGES RELAYCOUNTRY ASN noASNCIDR)],
 
 =item ignore_site_cf_files
 
@@ -506,17 +506,90 @@ supplementary attribute of the message, typically information that cannot
 be deduced from the message itself, or is hard to do so reliably, or would
 represent unnecessary work for SpamAssassin to obtain it. The argument will
 be stored to a Mail::SpamAssassin::Message object as 'suppl_attrib', thus
-made available to the rest of the code as well as to plugins. The exact list
-of attributes will evolve through time, any unknown attribute should be
-ignored. Possible examples are: SMTP envelope information, a flag indicating
-that a message as supplied by a caller was truncated due to size limit, an
-already verified list of DKIM signature objects, or perhaps a list of rule
-hits predetermined by a caller, which makes another possible way for a
-caller to provide meta information (instead of having to insert made-up
-header fields in order to pass information), or maybe just plain rule hits.
+made available to the rest of the code as well as to plugins. Possible
+attributes are:
 
-For more information, please see the C<Mail::SpamAssassin::Message>
-and C<Mail::SpamAssassin::Message::Node> POD.
+=over 4
+
+=item arc_signatures
+
+An array reference of ARC signatures. If this attribute is provided,
+SpamAssassin will not attempt to verify ARC signatures itself, but will use
+the provided signatures instead. This is useful when the caller has already
+verified the signatures and wants to avoid the overhead of verifying them
+again or if the message was truncated before being passed to SpamAssassin.
+
+=item body_size
+
+The original message body size. The caller may wish to provide this value
+if the message was truncated before being passed to SpamAssassin. The value
+will be stored as I<< $msg->{pristine_body_length} >> and used by an eval test
+C<check_body_length>
+
+=item dkim_signatures
+
+An array reference of DKIM signatures. If this attribute is provided,
+SpamAssassin will not attempt to verify DKIM signatures itself, but will use
+the provided signatures instead. This is useful when the caller has already
+verified the signatures and wants to avoid the overhead of verifying them
+again or if the message was truncated before being passed to SpamAssassin.
+
+=item master_deadline
+
+This attribute is used to set a limit on the time SpamAssassin is allowed to
+spend before retuning a result. The value is a floating point number
+representing the time in seconds since the epoch. See the C<time_limit>
+configuration option in L<Mail::SpamAssassin::Conf/MISCELLANEOUS OPTIONS>
+for more details.
+
+=item mimepart_digests
+
+An array reference of digest codes (e.g. SHA1) of each MIME part. The caller may wish
+to provide this value if the message was truncated before being passed to
+SpamAssassin.
+
+=item originating
+
+A boolean value that indicates whether the message originated from the
+local system or, likewise, an authenticated roaming user. If this is set to true,
+SpamAssassin will consider all relays as trusted.
+
+=item return_path
+
+This attribute is used to set the Envelope From address that is used by the
+C<EnvelopeFrom> pseudo-header and for various rules such as C<welcomelist_from>
+and SPF checking. If this attribute is not provided, SpamAssassin will attempt to
+determine the Envelope From address from the message headers. If the caller knows
+the Envelope From address, it can be provided here to avoid any ambiguity. If the
+address is surrounded by angle brackets, they will be stripped.
+
+=item rule_hits
+
+A array reference of rule hits. Each element is a hash reference with the
+following keys: 'rule', 'area', 'score', 'defscore', 'value', 'ruletype', 'tflags',
+'descr'. This is useful when the caller wants to include additional hits on the
+message that are not defined by SpamAssassin rules. For example, this can be used
+to include hits from external sources such as virus scanners or other spam filters.
+
+=back
+
+Example:
+
+    $msg = $sa->parse($message,0, {
+        body_size       => 24284293,
+        master_deadline => time() + 60,
+        return_path     => 'foo@example.com',
+        originating     => 0,
+        rule_hits       => {
+            rule   => '__TRUNCATED',
+            score  => -0.1,
+            area   => 'RAW: ',
+            tflags => 'nice',
+            descr  => "Message size truncated to 10485760 B" }
+    });
+
+If other attributes are provided, they will be silently ignored by SpamAssassin but
+will be available to plugins in C<< $msg->{suppl_attrib} >>.
 
 =cut
 
@@ -561,7 +634,7 @@ sub parse {
 
 ###########################################################################
 
-=item $status = $f->check ($mail)
+=item $status = $f-E<gt>check ($mail)
 
 Check a mail, encapsulated in a C<Mail::SpamAssassin::Message> object,
 to determine if it is spam or not.
@@ -586,7 +659,7 @@ sub check {
   $pms;
 }
 
-=item $status = $f->check_message_text ($mailtext)
+=item $status = $f-E<gt>check_message_text ($mailtext)
 
 Check a mail, encapsulated in a plain string C<$mailtext>, to determine if it
 is spam or not.
@@ -612,7 +685,7 @@ sub check_message_text {
 
 ###########################################################################
 
-=item $status = $f->learn ($mail, $id, $isspam, $forget)
+=item $status = $f-E<gt>learn ($mail, $id, $isspam, $forget)
 
 Learn from a mail, encapsulated in a C<Mail::SpamAssassin::Message> object.
 
@@ -663,7 +736,7 @@ sub learn {
 
 ###########################################################################
 
-=item $f->init_learner ( [ { opt => val, ... } ] )
+=item $f-E<gt>init_learner ( [ { opt =E<gt> val, ... } ] )
 
 Initialise learning.  You may pass the following attribute-value pairs to this
 method.
@@ -735,7 +808,7 @@ sub init_learner {
 
 ###########################################################################
 
-=item $f->rebuild_learner_caches ({ opt => val })
+=item $f-E<gt>rebuild_learner_caches ({ opt =E<gt> val })
 
 Rebuild any cache databases; should be called after the learning process.
 Options include: C<verbose>, which will output diagnostics to C<stdout>
@@ -750,7 +823,7 @@ sub rebuild_learner_caches {
   1;
 }
 
-=item $f->finish_learner ()
+=item $f-E<gt>finish_learner ()
 
 Finish learning.
 
@@ -763,7 +836,7 @@ sub finish_learner {
   1;
 }
 
-=item $f->dump_bayes_db()
+=item $f-E<gt>dump_bayes_db()
 
 Dump the contents of the Bayes DB
 
@@ -774,7 +847,7 @@ sub dump_bayes_db {
   $self->{bayes_scanner}->dump_bayes_db(@opts) if $self->{bayes_scanner};
 }
 
-=item $f->signal_user_changed ( [ { opt => val, ... } ] )
+=item $f-E<gt>signal_user_changed ( [ { opt =E<gt> val, ... } ] )
 
 Signals that the current user has changed (possibly using C<setuid>), meaning
 that SpamAssassin should close any per-user databases it has open, and re-open
@@ -853,13 +926,13 @@ sub signal_user_changed {
 		userstate_dir => $self->{userstate_dir},
 		user_dir => $self->{user_dir},
 	      });
-
+  undef $self->{cf_files_read};
   1;
 }
 
 ###########################################################################
 
-=item $f->report_as_spam ($mail, $options)
+=item $f-E<gt>report_as_spam ($mail, $options)
 
 Report a mail, encapsulated in a C<Mail::SpamAssassin::Message> object, as
 human-verified spam.  This will submit the mail message to live,
@@ -912,7 +985,7 @@ sub report_as_spam {
 
 ###########################################################################
 
-=item $f->revoke_as_spam ($mail, $options)
+=item $f-E<gt>revoke_as_spam ($mail, $options)
 
 Revoke a mail, encapsulated in a C<Mail::SpamAssassin::Message> object, as
 human-verified ham (non-spam).  This will revoke the mail message from live,
@@ -952,7 +1025,7 @@ sub revoke_as_spam {
 
 ###########################################################################
 
-=item $f->add_address_to_welcomelist ($addr, $cli_p)
+=item $f-E<gt>add_address_to_welcomelist ($addr, $cli_p)
 
 Previously add_address_to_whitelist which will work interchangeably until 4.1.
 
@@ -973,7 +1046,7 @@ sub add_address_to_welcomelist {
 
 ###########################################################################
 
-=item $f->add_all_addresses_to_welcomelist ($mail, $cli_p)
+=item $f-E<gt>add_all_addresses_to_welcomelist ($mail, $cli_p)
 
 Previously add_all_addresses_to_whitelist which will work interchangeably until 4.1.
 
@@ -996,7 +1069,7 @@ sub add_all_addresses_to_welcomelist {
 
 ###########################################################################
 
-=item $f->remove_address_from_welcomelist ($addr, $cli_p)
+=item $f-E<gt>remove_address_from_welcomelist ($addr, $cli_p)
 
 Previously remove_address_from_whitelist which will work interchangeably until 4.1.
 
@@ -1013,11 +1086,12 @@ sub remove_address_from_welcomelist {
   $self->call_plugins("remove_address", { address => $addr,
                                           cli_p => $cli_p });
 }
+*remove_address_from_blocklist = \&remove_address_from_welcomelist;
 *remove_address_from_whitelist = \&remove_address_from_welcomelist; # removed in 4.1
 
 ###########################################################################
 
-=item $f->remove_all_addresses_from_welcomelist ($mail, $cli_p)
+=item $f-E<gt>remove_all_addresses_from_welcomelist ($mail, $cli_p)
 
 Previously remove_all_addresses_from_whitelist which will work interchangeably until 4.1.
 
@@ -1041,7 +1115,7 @@ sub remove_all_addresses_from_welcomelist {
 
 ###########################################################################
 
-=item $f->add_address_to_blocklist ($addr, $cli_p)
+=item $f-E<gt>add_address_to_blocklist ($addr, $cli_p)
 
 Previously add_address_to_blacklist which will work interchangeably until 4.1.
 
@@ -1061,7 +1135,7 @@ sub add_address_to_blocklist {
 
 ###########################################################################
 
-=item $f->add_all_addresses_to_blocklist ($mail, $cli_p)
+=item $f-E<gt>add_all_addresses_to_blocklist ($mail, $cli_p)
 
 Previously add_all_addresses_to_blacklist which will work interchangeably until 4.1.
 
@@ -1097,7 +1171,7 @@ sub add_all_addresses_to_blocklist {
 
 ###########################################################################
 
-=item $text = $f->remove_spamassassin_markup ($mail)
+=item $text = $f-E<gt>remove_spamassassin_markup ($mail)
 
 Returns the text of the message, with any SpamAssassin-added text (such
 as the report, or X-Spam-Status headers) stripped.
@@ -1249,7 +1323,7 @@ sub remove_spamassassin_markup {
 
 ###########################################################################
 
-=item $f->read_scoreonly_config ($filename)
+=item $f-E<gt>read_scoreonly_config ($filename)
 
 Read a configuration file and parse user preferences from it.
 
@@ -1292,7 +1366,7 @@ sub read_scoreonly_config {
 
 ###########################################################################
 
-=item $f->load_scoreonly_sql ($username)
+=item $f-E<gt>load_scoreonly_sql ($username)
 
 Read configuration parameters from SQL database and parse scores from it.  This
 will only take effect if the perl C<DBI> module is installed, and the
@@ -1318,7 +1392,7 @@ sub load_scoreonly_sql {
 
 ###########################################################################
 
-=item $f->load_scoreonly_ldap ($username)
+=item $f-E<gt>load_scoreonly_ldap ($username)
 
 Read configuration parameters from an LDAP server and parse scores from it.
 This will only take effect if the perl C<Net::LDAP> and C<URI> modules are
@@ -1343,7 +1417,7 @@ sub load_scoreonly_ldap {
 
 ###########################################################################
 
-=item $f->set_persistent_address_list_factory ($factoryobj)
+=item $f-E<gt>set_persistent_address_list_factory ($factoryobj)
 
 Set the persistent address list factory, used to create objects for the
 automatic welcomelist algorithm's persistent-storage back-end.  See
@@ -1359,7 +1433,7 @@ sub set_persistent_address_list_factory {
 
 ###########################################################################
 
-=item $f->compile_now ($use_user_prefs, $keep_userstate)
+=item $f-E<gt>compile_now ($use_user_prefs, $keep_userstate)
 
 Compile all patterns, load all configuration files, and load all
 possibly-required Perl modules.
@@ -1471,7 +1545,7 @@ sub compile_now {
 
 ###########################################################################
 
-=item $f->debug_diagnostics ()
+=item $f-E<gt>debug_diagnostics ()
 
 Output some diagnostic information, useful for debugging SpamAssassin
 problems.
@@ -1491,7 +1565,7 @@ sub debug_diagnostics {
 
 ###########################################################################
 
-=item $failed = $f->lint_rules ()
+=item $failed = $f-E<gt>lint_rules ()
 
 Syntax-check the current set of rules.  Returns the number of 
 syntax errors discovered, or 0 if the configuration is valid.
@@ -1536,7 +1610,7 @@ sub lint_rules {
 
 ###########################################################################
 
-=item $f->finish()
+=item $f-E<gt>finish()
 
 Destroy this object, so that it will be garbage-collected once it
 goes out of scope.  The object will no longer be usable after this
@@ -2017,7 +2091,7 @@ sub test_global_state_dir {
     return 0;
 }
 
-=item $fullpath = $f->find_rule_support_file ($filename)
+=item $fullpath = $f-E<gt>find_rule_support_file ($filename)
 
 Find a rule-support file, such as C<languages> or C<triplets.txt>,
 in the system-wide rules directory, and return its full path if
@@ -2048,7 +2122,7 @@ sub find_rule_support_file {
     map { my $p = $_; $p =~ s{$}{/$filename}; $p } @paths );
 }
 
-=item $f->create_default_prefs ($filename, $username [ , $userdir ] )
+=item $f-E<gt>create_default_prefs ($filename, $username [ , $userdir ] )
 
 Copy default preferences file into home directory for later use and
 modification, if it does not already exist and C<dont_copy_prefs> is
@@ -2374,7 +2448,7 @@ sub sa_die {
 
 ###########################################################################
 
-=item $f->copy_config ( [ $source ], [ $dest ] )
+=item $f-E<gt>copy_config ( [ $source ], [ $dest ] )
 
 Used for daemons to keep a persistent Mail::SpamAssassin object's
 configuration correct if switching between users.  Pass an associative
@@ -2425,7 +2499,7 @@ sub copy_config {
 
 ###########################################################################
 
-=item @plugins = $f->get_loaded_plugins_list ( )
+=item @plugins = $f-E<gt>get_loaded_plugins_list ( )
 
 Return the list of plugins currently loaded by this SpamAssassin object's
 configuration; each entry in the list is an object of type
@@ -2466,7 +2540,7 @@ sa-update(1)
 
 =head1 BUGS
 
-See E<lt>https://issues.apache.org/SpamAssassin/E<gt>
+See E<lt>https://bz.apache.org/SpamAssassin/E<gt>
 
 =head1 AUTHORS
 
