@@ -39,7 +39,7 @@
 # (e.g. from cymru.com). This form of a response is handled as well.
 #
 # Some zones also support IPv6 lookups, for example:
-#   asn_lookup origin6.asn.cymru.com [_ASN_ _ASNCIDR_]
+#   asn_lookup_ipv6 origin6.asn.cymru.com [_ASN_ _ASNCIDR_]
 
 =head1 NAME
 
@@ -50,26 +50,52 @@ Autonomous System Number (ASN) of the connecting IP address.
 
  loadplugin Mail::SpamAssassin::Plugin::ASN
 
- asn_lookup asn.routeviews.org _ASN_ _ASNCIDR_
+ # Default / recommended settings
+ asn_use_geodb 1
+ asn_use_dns 1
+ asn_prefer_geodb 1
 
+ # Do lookups and add tags / X-Spam-ASN header
+ asn_lookup asn.routeviews.org _ASN_ _ASNCIDR_
+ asn_lookup_ipv6 origin6.asn.cymru.com _ASN_ _ASNCIDR_
  add_header all ASN _ASN_ _ASNCIDR_
 
+ # Rules to test ASN or Organization
+ # NOTE: Do not use rules that check metadata X-ASN header,
+ # only check_asn() eval function works correctly.
+ # Rule argument is full regexp to match.
+
+ # ASN Number: GeoIP ASN or DNS
+ # Matched string includes asn_prefix if defined, and normally
+ # looks like "AS1234" (DNS) or "AS1234 Google LLC" (GeoIP)
+ header AS_1234 eval:check_asn('/^AS1234\b/')
+
+ # ASN Organisation: GeoIP ASN has, DNS lists might not have
+ # Note the second parameter which checks MYASN tag (default is ASN)
+ asn_lookup myview.example.com _MYASN_ _MYASNCIDR_
+ header AS_GOOGLE eval:check_asn('/\bGoogle\b/i', 'MYASN')
+
 =head1 DESCRIPTION
 
 This plugin uses DNS lookups to the services of an external DNS zone such
-as at C<http://www.routeviews.org/> to do the actual work. Please make
+as at C<https://www.routeviews.org/> to do the actual work. Please make
 sure that your use of the plugin does not overload their infrastructure -
 this generally means that B<you should not use this plugin in a
 high-volume environment> or that you should use a local mirror of the
 zone (see C<ftp://ftp.routeviews.org/dnszones/>).  Other similar zones
 may also be used.
 
+GeoDB (GeoIP ASN) database lookups are supported since SpamAssassin 4.0 and
+it's recommended to use them instead of DNS queries, unless C<_ASNCIDR_>
+is needed.
+
 =head1 TEMPLATE TAGS
 
 This plugin allows you to create template tags containing the connecting
 IP's AS number and route info for that AS number.
 
-The default config will add a header field that looks like this:
+If you use add_header as documented in the example before, a header field is
+added that looks like this:
 
  X-Spam-ASN: AS24940 213.239.192.0/18
 
@@ -83,69 +109,35 @@ all be added to the C<_ASNCIDR_> tag, separated by spaces, eg:
 Note that the literal "AS" before the ASN in the _ASN_ tag is configurable
 through the I<asn_prefix> directive and may be set to an empty string.
 
-=head1 CONFIGURATION
+C<_ASNCIDR_> is not available with local GeoDB ASN lookups.
 
-The standard ruleset contains a configuration that will add a header field
-containing ASN data to scanned messages.  The bayes tokenizer will use the
-added header field for bayes calculations, and thus affect which BAYES_* rule
-will trigger for a particular message.
+=head1 USER SETTINGS
 
-B<Note> that in most cases you should not score on the ASN data directly.
-Bayes learning will probably trigger on the _ASNCIDR_ tag, but probably not
-very well on the _ASN_ tag alone.
+=over 4
 
-=head1 SEE ALSO
+=item clear_asn_lookups
 
-http://www.routeviews.org/ - all data regarding routing, ASNs, etc....
+Removes all previously declared I<asn_lookup> or I<asn_lookup_ipv6> entries
+from the list of queries.
 
-http://issues.apache.org/SpamAssassin/show_bug.cgi?id=4770 -
-SpamAssassin Issue #4770 concerning this plugin
+=item asn_prefix 'prefix_string'       (default: 'AS')
 
-=head1 STATUS
+The string specified in the argument is prepended to each ASN when storing
+it as a tag. This prefix is rather redundant, but its default value 'AS'
+is kept for backward compatibility with versions of SpamAssassin earlier
+than 3.4.0. A sensible setting is an empty string. The argument may be (but
+need not be) enclosed in single or double quotes for clarity.
 
-No in-depth analysis of the usefulness of bayes tokenization of ASN data has
-been performed.
+=back
 
-=cut
-
-package Mail::SpamAssassin::Plugin::ASN;
-
-use strict;
-use warnings;
-use re 'taint';
-use Mail::SpamAssassin;
-use Mail::SpamAssassin::Plugin;
-use Mail::SpamAssassin::Logger;
-use Mail::SpamAssassin::Util qw(reverse_ip_address);
-use Mail::SpamAssassin::Dns;
-
-our @ISA = qw(Mail::SpamAssassin::Plugin);
-
-sub new {
-  my ($class, $mailsa) = @_;
-  $class = ref($class) || $class;
-  my $self = $class->SUPER::new($mailsa);
-  bless ($self, $class);
-
-  $self->set_config($mailsa->{conf});
-
-  return $self;
-}
-
-###########################################################################
-
-sub set_config {
-  my ($self, $conf) = @_;
-  my @cmds;
-
-=head1 ADMINISTRATOR SETTINGS
+=head1 RULE DEFINITIONS AND PRIVILEGED SETTINGS
 
 =over 4
 
 =item asn_lookup asn-zone.example.com [ _ASNTAG_ _ASNCIDRTAG_ ]
 
 Use this to lookup the ASN info in the specified zone for the first external
-IP address and add the AS number to the first specified tag and routing info
+IPv4 address and add the AS number to the first specified tag and routing info
 to the second specified tag.
 
 If no tags are specified the AS number will be added to the _ASN_ tag and the
@@ -170,33 +162,90 @@ Examples:
 
   asn_lookup in1tag.example.net _ASNDATA_ _ASNDATA_
 
+=item asn_lookup_ipv6 asn-zone6.example.com [_ASN_ _ASNCIDR_]
+
+Use specified zone for lookups of IPv6 addresses.  If zone supports both
+IPv4 and IPv6 queries, use both asn_lookup and asn_lookup_ipv6 for the same
+zone.
+
 =back
 
+=head1 ADMINISTRATOR SETTINGS
+
 =over 4
 
-=item clear_asn_lookups
+=item asn_use_geodb ( 0 / 1 )          (default: 1)
+
+Use Mail::SpamAssassin::GeoDB module to lookup ASN numbers.  You need
+suitable supported module like GeoIP2 or GeoIP with ISP or ASN database
+installed (for example, add EditionIDs GeoLite2-ASN in GeoIP.conf for
+geoipupdate program).
+
+GeoDB can only set _ASN_ tag, it has no data for _ASNCIDR_.  If you need
+both, then set asn_prefer_geodb 0 so DNS rules are tried.
+
+=item asn_prefer_geodb ( 0 / 1 )       (default: 1)
+
+If set, DNS lookups (asn_lookup rules) will not be run if GeoDB successfully
+finds ASN. Set this to 0 to get _ASNCIDR_ even if GeoDB finds _ASN_.
+
+=item asn_use_dns ( 0 / 1 )            (default: 1)
+
+Set to 0 to never allow DNS queries.
 
 =back
 
-Removes any previously declared I<asn_lookup> entries from a list of queries.
+=head1 BAYES
 
-=over 4
+The bayes tokenizer will use ASN data for bayes calculations, and thus
+affect which BAYES_* rule will trigger for a particular message.  No
+in-depth analysis of the usefulness of bayes tokenization of ASN data has
+been performed.
 
-=item asn_prefix 'prefix_string'       (default: 'AS')
+=head1 SEE ALSO
 
-The string specified in the argument is prepended to each ASN when storing
-it as a tag. This prefix is rather redundant, but its default value 'AS'
-is kept for backward compatibility with versions of SpamAssassin earlier
-than 3.4.0. A sensible setting is an empty string. The argument may be (but
-need not be) enclosed in single or double quotes for clarity.
-
-=back
+https://www.routeviews.org/ - all data regarding routing, ASNs, etc....
 
 =cut
 
+package Mail::SpamAssassin::Plugin::ASN;
+
+use strict;
+use warnings;
+use re 'taint';
+
+use Mail::SpamAssassin::Plugin;
+use Mail::SpamAssassin::Logger;
+use Mail::SpamAssassin::Util qw(reverse_ip_address compile_regexp);
+use Mail::SpamAssassin::Constants qw(:ip);
+
+our @ISA = qw(Mail::SpamAssassin::Plugin);
+
+sub new {
+  my ($class, $mailsa) = @_;
+  $class = ref($class) || $class;
+  my $self = $class->SUPER::new($mailsa);
+  bless ($self, $class);
+
+  $self->register_eval_rule("check_asn", $Mail::SpamAssassin::Conf::TYPE_HEAD_EVALS);
+
+  $self->set_config($mailsa->{conf});
+
+  # we need GeoDB ASN
+  $self->{main}->{geodb_wanted}->{asn} = 1;
+
+  return $self;
+}
+
+###########################################################################
+
+sub set_config {
+  my ($self, $conf) = @_;
+  my @cmds;
+
   push (@cmds, {
     setting => 'asn_lookup',
-    is_admin => 1,
+    is_priv => 1,
     code => sub {
       my ($conf, $key, $value, $line) = @_;
       unless (defined $value && $value !~ /^$/) {
@@ -214,9 +263,28 @@ need not be) enclosed in single or double quotes for clarity.
     }
   });
 
+  push (@cmds, {
+    setting => 'asn_lookup_ipv6',
+    is_priv => 1,
+    code => sub {
+      my ($conf, $key, $value, $line) = @_;
+      unless (defined $value && $value !~ /^$/) {
+        return $Mail::SpamAssassin::Conf::MISSING_REQUIRED_VALUE;
+      }
+      local($1,$2,$3);
+      unless ($value =~ /^(\S+?)\.?(?:\s+_(\S+)_\s+_(\S+)_)?$/) {
+        return $Mail::SpamAssassin::Conf::INVALID_VALUE;
+      }
+      my ($zone, $asn_tag, $route_tag) = ($1, $2, $3);
+      $asn_tag   = 'ASN'     if !defined $asn_tag;
+      $route_tag = 'ASNCIDR' if !defined $route_tag;
+      push @{$conf->{asnlookups_ipv6}},
+           { zone=>$zone, asn_tag=>$asn_tag, route_tag=>$route_tag };
+    }
+  });
+
   push (@cmds, {
     setting => 'clear_asn_lookups',
-    is_admin => 1,
     type => $Mail::SpamAssassin::Conf::CONF_TYPE_NOARGS,
     code => sub {
       my ($conf, $key, $value, $line) = @_;
@@ -224,6 +292,7 @@ need not be) enclosed in single or double quotes for clarity.
         return $Mail::SpamAssassin::Conf::INVALID_VALUE;
       }
       delete $conf->{asnlookups};
+      delete $conf->{asnlookups_ipv6};
     }
   });
 
@@ -240,72 +309,154 @@ need not be) enclosed in single or double quotes for clarity.
     }
   });
 
+  push (@cmds, {
+    setting => 'asn_use_geodb',
+    default => 1,
+    is_admin => 1,
+    type => $Mail::SpamAssassin::Conf::CONF_TYPE_BOOL,
+  });
+
+  push (@cmds, {
+    setting => 'asn_prefer_geodb',
+    default => 1,
+    is_admin => 1,
+    type => $Mail::SpamAssassin::Conf::CONF_TYPE_BOOL,
+  });
+
+  push (@cmds, {
+    setting => 'asn_use_dns',
+    default => 1,
+    is_admin => 1,
+    type => $Mail::SpamAssassin::Conf::CONF_TYPE_BOOL,
+  });
+
   $conf->{parser}->register_commands(\@cmds);
 }
 
 # ---------------------------------------------------------------------------
 
-sub parsed_metadata {
+sub extract_metadata {
   my ($self, $opts) = @_;
 
   my $pms = $opts->{permsgstatus};
-  my $conf = $self->{main}->{conf};
+  my $conf = $pms->{conf};
 
-  unless ($conf->{asnlookups}) {
-    dbg("asn: no asn_lookup configured, skipping ASN lookups");
-    return; # no asn_lookups mean no tags need to be initialized
+  my $geodb = $self->{main}->{geodb};
+  my $has_geodb = $conf->{asn_use_geodb} && $geodb && $geodb->can('asn');
+  if ($has_geodb) {
+    dbg("asn: using GeoDB ASN for lookups");
+  } else {
+    dbg("asn: GeoDB ASN not available");
+    if (!$conf->{asn_use_dns} || !$pms->is_dns_available()) {
+      dbg("asn: DNS is not available, skipping ASN check");
+      return;
+    }
+    if ($self->{main}->{learning}) {
+      dbg("asn: learning message, skipping DNS-based ASN check");
+      return;
+    }
   }
 
-  # get reversed IP address of last external relay to lookup
-  # don't return until we've initialized the template tags
-  my($ip,$reversed_ip);
-  my $relay = $pms->{relays_external}->[0];
-  $ip = $relay->{ip}  if defined $relay;
-  if (!$pms->is_dns_available()) {
-    dbg("asn: DNS is not available, skipping ASN checks");
-  } elsif (!defined $ip) {
+  # initialize the tag data so that if no result is returned from the DNS
+  # query we won't end up with a missing tag.  Don't use $pms->set_tag()
+  # here to avoid triggering any tag-dependent action unnecessarily
+  if ($conf->{asnlookups}) {
+    foreach my $entry (@{$conf->{asnlookups}}) {
+      $pms->{tag_data}->{$entry->{asn_tag}} ||= '';
+      $pms->{tag_data}->{$entry->{route_tag}} ||= '';
+    }
+  }
+  if ($conf->{asnlookups_ipv6}) {
+    foreach my $entry (@{$conf->{asnlookups_ipv6}}) {
+      $pms->{tag_data}->{$entry->{asn_tag}} ||= '';
+      $pms->{tag_data}->{$entry->{route_tag}} ||= '';
+    }
+  }
+
+  # Initialize status
+  $pms->{asn_results} = ();
+
+  # get IP address of last external relay to lookup
+  my $relay = $opts->{msg}->{metadata}->{relays_external}->[0];
+  if (!defined $relay) {
     dbg("asn: no first external relay IP available, skipping ASN check");
+    return;
   } elsif ($relay->{ip_private}) {
     dbg("asn: first external relay is a private IP, skipping ASN check");
-  } else {
-    $reversed_ip = reverse_ip_address($ip);
-    if (defined $reversed_ip) {
-      dbg("asn: using first external relay IP for lookups: %s", $ip);
+    return;
+  }
+  my $ip = $relay->{ip};
+  dbg("asn: using first external relay IP for lookups: %s", $ip);
+
+  # GeoDB lookup
+  my $asn_found;
+  if ($has_geodb) {
+    my $asn = $geodb->get_asn($ip);
+    my $org = $geodb->get_asn_org($ip);
+    if (!defined $asn) {
+      dbg("asn: GeoDB ASN lookup failed");
     } else {
-      dbg("asn: could not parse first external relay IP: %s, skipping", $ip);
+      $asn_found = 1;
+      dbg("asn: GeoDB found ASN $asn");
+      # Prevent double prefix
+      my $asn_value =
+        length($conf->{asn_prefix}) && index($asn, $conf->{asn_prefix}) != 0 ?
+          $conf->{asn_prefix}.$asn : $asn;
+      $asn_value .= ' '.$org if defined $org && length($org);
+      $pms->set_tag('ASN', $asn_value);
+      # For Bayes
+      $pms->{msg}->put_metadata('X-ASN', $asn);
     }
   }
 
+  # Skip DNS if GeoDB was successful and preferred
+  if ($asn_found && $conf->{asn_prefer_geodb}) {
+    dbg("asn: GeoDB lookup successful, skipping DNS lookups");
+    return;
+  }
+
+  # No point continuing without DNS from now on
+  if (!$conf->{asn_use_dns} || !$pms->is_dns_available()) {
+    dbg("asn: skipping disabled DNS lookups");
+    return;
+  }
+
+  dbg("asn: using DNS for lookups");
+  my $lookup_zone;
+  if ($ip =~ IS_IPV4_ADDRESS) {
+    if (!defined $conf->{asnlookups}) {
+      dbg("asn: asn_lookup for IPv4 not defined, skipping");
+      return;
+    }
+    $lookup_zone = "asnlookups";
+  } else {
+    if (!defined $conf->{asnlookups_ipv6}) {
+      dbg("asn: asn_lookup_ipv6 for IPv6 not defined, skipping");
+      return;
+    }
+    $lookup_zone = "asnlookups_ipv6";
+  }
+  
+  my $reversed_ip = reverse_ip_address($ip);
+  if (!defined $reversed_ip) {
+    dbg("asn: could not parse IP: %s, skipping", $ip);
+    return;
+  }
+
   # we use arrays and array indices rather than hashes and hash keys
   # in case someone wants the same zone added to multiple sets of tags
   my $index = 0;
-  foreach my $entry (@{$conf->{asnlookups}}) {
-    # initialize the tag data so that if no result is returned from the DNS
-    # query we won't end up with a missing tag.  Don't use $pms->set_tag()
-    # here to avoid triggering any tag-dependent action unnecessarily
-    unless (defined $pms->{tag_data}->{$entry->{asn_tag}}) {
-      $pms->{tag_data}->{$entry->{asn_tag}} = '';
-    }
-    unless (defined $pms->{tag_data}->{$entry->{route_tag}}) {
-      $pms->{tag_data}->{$entry->{route_tag}} = '';
-    }
-    next unless $reversed_ip;
-
+  foreach my $entry (@{$conf->{$lookup_zone}}) {
     # do the DNS query, have the callback process the result
     my $zone_index = $index;
     my $zone = $reversed_ip . '.' . $entry->{zone};
-    my $key = "asnlookup-${zone_index}-$entry->{zone}";
-    my $ent = $pms->{async}->bgsend_and_start_lookup(
-        $zone, 'TXT', undef,
-        { key => $key, zone => $zone },
-        sub { my($ent, $pkt) = @_;
-              $self->process_dns_result($pms, $pkt, $zone_index) },
-      master_deadline => $pms->{master_deadline} );
-    if ($ent) {
-      dbg("asn: launched DNS TXT query for %s.%s in background",
-          $reversed_ip, $entry->{zone});
-      $index++;
-    }
+    $pms->{async}->bgsend_and_start_lookup($zone, 'TXT', undef,
+      { rulename => 'asn_lookup', type => 'ASN' },
+      sub { my($ent, $pkt) = @_;
+            $self->process_dns_result($pms, $pkt, $zone_index, $lookup_zone) },
+      master_deadline => $pms->{master_deadline}
+    );
+    $index++;
   }
 }
 
@@ -320,13 +471,16 @@ sub parsed_metadata {
 #       be no CIDR field in that case.
 #
 sub process_dns_result {
-  my ($self, $pms, $pkt, $zone_index) = @_;
+  my ($self, $pms, $pkt, $zone_index, $lookup_zone) = @_;
+
+  # NOTE: $pkt will be undef if the DNS query was aborted (e.g. timed out)
+  return if !$pkt;
 
   my $conf = $self->{main}->{conf};
 
-  my $zone = $conf->{asnlookups}[$zone_index]->{zone};
-  my $asn_tag = $conf->{asnlookups}[$zone_index]->{asn_tag};
-  my $route_tag = $conf->{asnlookups}[$zone_index]->{route_tag};
+  my $zone = $conf->{$lookup_zone}[$zone_index]->{zone};
+  my $asn_tag = $conf->{$lookup_zone}[$zone_index]->{asn_tag};
+  my $route_tag = $conf->{$lookup_zone}[$zone_index]->{route_tag};
 
   my($any_asn_updates, $any_route_updates, $tag_value);
 
@@ -349,14 +503,12 @@ sub process_dns_result {
     %route_tag_data_seen = map(($_,1), @route_tag_data);
   }
 
-  # NOTE: $pkt will be undef if the DNS query was aborted (e.g. timed out)
-  my @answer = !defined $pkt ? () : $pkt->answer;
-
-  foreach my $rr (@answer) {
-    dbg("asn: %s: lookup result packet: %s", $zone, $rr->string);
+  foreach my $rr ($pkt->answer) {
+    #dbg("asn: %s: lookup result packet: %s", $zone, $rr->string);
     next if $rr->type ne 'TXT';
-    my @strings = $rr->char_str_list;
+    my @strings = $rr->txtdata;
     next if !@strings;
+    for (@strings) { utf8::encode($_) if utf8::is_utf8($_) }
 
     my @items;
     if (@strings > 1 && join('',@strings) !~ m{\|}) {
@@ -422,10 +574,56 @@ sub process_dns_result {
                   @asn_tag_data == 1 ? $asn_tag_data[0] : \@asn_tag_data);
   }
   if ($any_route_updates && @route_tag_data) {
-    $pms->{msg}->put_metadata('X-ASN-Route', join(' ',@route_tag_data));
+    # Bayes already has X-ASN, Route is pointless duplicate, skip
+    #$pms->{msg}->put_metadata('X-ASN-Route', join(' ',@route_tag_data));
     $pms->set_tag($route_tag,
                   @route_tag_data == 1 ? $route_tag_data[0] : \@route_tag_data);
   }
 }
 
+sub check_asn {
+  my ($self, $pms, $re, $asn_tag) = @_;
+
+  my $rulename = $pms->get_current_eval_rule_name();
+  if (!defined $re) {
+    warn "asn: rule $rulename eval argument missing\n";
+    return 0;
+  }
+
+  my ($rec, $err) = compile_regexp($re, 2);
+  if (!$rec) {
+    warn "asn: invalid regexp for $rulename '$re': $err\n";
+    return 0;
+  }
+
+  $asn_tag = 'ASN' unless defined $asn_tag;
+  $pms->action_depends_on_tags($asn_tag,
+    sub { my($pms,@args) = @_;
+      $self->_check_asn($pms, $rulename, $rec, $asn_tag);
+    }
+  );
+
+  return; # return undef for async status
+}
+
+sub _check_asn {
+  my ($self, $pms, $rulename, $rec, $asn_tag) = @_;
+
+  $pms->rule_ready($rulename); # mark rule ready for metas
+
+  my $asn = $pms->get_tag($asn_tag);
+  return if !defined $asn;
+
+  if ($asn =~ $rec) {
+    $pms->test_log("$asn_tag: $asn", $rulename);
+    $pms->got_hit($rulename, "");
+  }
+}
+
+# Version features
+sub has_asn_lookup_ipv6 { 1 }
+sub has_asn_geodb { 1 }
+sub has_check_asn { 1 }
+sub has_check_asn_tag { 1 } # $asn_tag parameter for check_asn()
+
 1;
