@@ -33,21 +33,10 @@ responses trickle in, filters them according to the requested DNS resource
 record type and optional subrule filtering expression, yielding a rule hit
 if a response meets filtering conditions.
 
-=head1 USER SETTINGS
+Any host or its domain part matching uridnsbl_skip_domains is ignored
+by default.
 
-=over 4
-
-=item rbl_timeout t [t_min] [zone]		(default: 15 3)
-
-The rbl_timeout setting is common to all DNS querying rules (as implemented
-by other plugins). It can specify a DNS query timeout globally, or individually
-for each zone. When the zone parameter is specified, the settings affects DNS
-queries when their query domain equals the specified zone, or is its subdomain.
-See the C<Mail::SpamAssassin::Conf> POD for details on C<rbl_timeout>.
-
-=back
-
-=head1 RULE DEFINITIONS
+=head1 RULE DEFINITIONS AND PRIVILEGED SETTINGS
 
 =over 4
 
@@ -56,18 +45,32 @@ See the C<Mail::SpamAssassin::Conf> POD for details on C<rbl_timeout>.
 A query template is a string which will be expanded to produce a domain name
 to be used in a DNS query. The template may include SpamAssassin tag names,
 which will be replaced by their values to form a final query domain.
+
 The final query domain must adhere to rules governing DNS domains, i.e.
-must consist of fields each up to 63 characters long, delimited by dots.
-There may be a trailing dot at the end, but it is redundant / carries
-no semantics, because SpamAssassin uses a Net::DSN::Resolver::send method
-for querying DNS, which ignores any 'search' or 'domain' DNS resolver options.
+must consist of fields each up to 63 characters long, delimited by dots,
+not exceeding 255 characters. International domain names (in UTF-8) are
+allowed and will be encoded to ASCII-compatible encoding (ACE) according
+to IDN rules. Syntactically invalid resulting queries will be discarded
+by the DNS resolver code (with some info warnings).
+
+There may be a trailing dot at the end, but it is redundant / carries no
+semantics, because SpamAssassin uses a Net::DSN::Resolver::send method for
+querying DNS, which ignores any 'search' or 'domain' DNS resolver options.
 Domain names in DNS queries are case-insensitive.
 
 A tag name is a string of capital letters, preceded and followed by an
-underscore character. This syntax mirrors the add_header setting, except that
-tags cannot have parameters in parenthesis when used in askdns templates.
-Tag names may appear anywhere in the template - each queried DNS zone
-prescribes how a query should be formed.
+underscore character.  This syntax mirrors the add_header setting, except
+that tags cannot have parameters in parenthesis when used in askdns
+templates (exceptions found below).  Tag names may appear anywhere in the
+template - each queried DNS zone prescribes how a query should be formed.
+
+Special supported tag HEADER() can be used to extract any mail header
+field content by specifying a header field name with optional modifiers
+as an argument, like what is supported by header rules.  For example
+_HEADER(Reply-To:addr:domain)_ can be used to extract a trimmed domain
+part of an address found in a Reply-To header field.
+See Mail::SpamAssassin::Conf documentation about header rules
+and their modifiers.
 
 A query template may contain any number of tag names including none,
 although in the most common anticipated scenario exactly one tag name would
@@ -114,7 +117,8 @@ parameter will only act as a filter on a result.
 
 Currently recognized RR types in the rr_type parameter are: ANY, A, AAAA,
 MX, TXT, PTR, NAPTR, NS, SOA, CERT, CNAME, DNAME, DHCID, HINFO, MINFO,
-RP, HIP, IPSECKEY, KX, LOC, SRV, SSHFP, SPF.
+RP, HIP, IPSECKEY, KX, LOC, GPOS, SRV, OPENPGPKEY, SSHFP, SPF, TLSA, URI,
+CAA, CSYNC.
 
 https://www.iana.org/assignments/dns-parameters/dns-parameters.xml
 
@@ -135,7 +139,7 @@ the response.
 When a plain string is used as a filter, it must be enclosed in single or
 double quotes. For the rule to hit, the response must match the filtering
 string exactly, and a RR type of a response must match the query type.
-Typical use is an exact text string for TXT queries, or an exact quad-dotted
+Typical use is an exact text string for TXT queries, or an exact dotted-quad
 IPv4 address. In case of a TXT or SPF resource record which can return
 multiple character-strings (as defined in Section 3.3 of [RFC1035]), these
 strings are concatenated with no delimiters before comparing the result
@@ -155,15 +159,15 @@ with a specified filtering value and tested to fall within a 127.0.0.0/8
 network range - the rule hits if the result is nonzero:
 ((r & n) != 0) && ((r & 0xff000000) == 0x7f000000).  An example: 0x10 .
 
-A pair of numerical values (each a decimal, hexadecimal or quad-dotted)
+A pair of numerical values (each a decimal, hexadecimal or dotted-quad)
 delimited by a '-' specifies an IPv4 address range, and a pair of values
 delimited by a '/' specifies an IPv4 address followed by a bitmask. Again,
 this type of filtering expression is primarily intended with RR type-A
 DNS queries. The rule hits if the RR type matches, and the returned IP
-address falls within the specified range: (r >= n1 && r <= n2), or
+address falls within the specified range: (r E<gt>= n1 && r E<lt>= n2), or
 masked with a bitmask matches the specified value: (r & m) == (n & m) .
 
-As a shorthand notation, a single quad-dotted value is equivalent to
+As a shorthand notation, a single dotted-quad value is equivalent to
 a n-n form, i.e. it must match the returned value exactly with all its bits.
 
 Some typical examples of a numeric filtering parameter are: 127.0.1.2,
@@ -180,6 +184,11 @@ rcode indicates an error.  Example: [NXDOMAIN], or [FormErr,ServFail,4,5] .
 
 =back
 
+=head1 NOTES
+
+DNS timeout can be set with C<rbl_timeout> option.  See the
+C<Mail::SpamAssassin::Conf> POD for details on C<rbl_timeout>.
+
 =cut
 
 package Mail::SpamAssassin::Plugin::AskDNS;
@@ -189,8 +198,10 @@ use warnings;
 use re 'taint';
 
 use Mail::SpamAssassin::Plugin;
-use Mail::SpamAssassin::Util qw(decode_dns_question_entry);
+use Mail::SpamAssassin::Util qw(decode_dns_question_entry idn_to_ascii
+                                compile_regexp is_fqdn_valid);
 use Mail::SpamAssassin::Logger;
+use Mail::SpamAssassin::Constants qw(:ip);
 use version 0.77;
 
 our @ISA = qw(Mail::SpamAssassin::Plugin);
@@ -202,8 +213,6 @@ our %rcode_value = (  # https://www.iana.org/assignments/dns-parameters, RFC 619
   BADMODE => 19, BADNAME => 20, BADALG => 21, BADTRUNC => 22,
 );
 
-our $txtdata_can_provide_a_list;
-
 sub new {
   my($class,$sa_main) = @_;
 
@@ -213,10 +222,6 @@ sub new {
 
   $self->set_config($sa_main->{conf});
 
-  #$txtdata_can_provide_a_list = Net::DNS->VERSION >= 0.69;
-  #more robust version check from Damyan Ivanov - Bug 7095
-  $txtdata_can_provide_a_list = version->parse(Net::DNS->VERSION) >= version->parse('0.69');
-
   return $self;
 }
 
@@ -227,7 +232,7 @@ sub new {
 # numerical values, or as a bracketed and comma-separated list of DNS rcode
 # names or their numerical codes. Recognized numerical forms are: m, n1-n2,
 # or n/m, where n,n1,n2,m can be any of: decimal digits, 0x followed by
-# up to 8 hexadecimal digits, or an IPv4 address in quad-dotted notation.
+# up to 8 hexadecimal digits, or an IPv4 address in dotted-quad notation.
 # The argument is checked for syntax, undef is returned on syntax errors.
 # A string that looks like a regular expression is converted to a compiled
 # Regexp object and returned as a result. Otherwise, numeric components of
@@ -253,20 +258,14 @@ sub parse_and_canonicalize_subtest {
   my $result;
 
   local($1,$2,$3);
-  # modifiers /a, /d, /l, /u in suffix form were added with perl 5.13.10 (5.14)
-  # currently known modifiers are [msixoadlu], but let's not be too picky here
-  if (     $subtest =~ m{^       /  (.+) /  ([a-z]*) \z}xs) {
-    $result = $2 ne '' ? qr{(?$2)$1} : qr{$1};
-  } elsif ($subtest =~ m{^ m \s* \( (.+) \) ([a-z]*) \z}xs) {
-    $result = $2 ne '' ? qr{(?$2)$1} : qr{$1};
-  } elsif ($subtest =~ m{^ m \s* \[ (.+) \] ([a-z]*) \z}xs) {
-    $result = $2 ne '' ? qr{(?$2)$1} : qr{$1};
-  } elsif ($subtest =~ m{^ m \s* \{ (.+) \} ([a-z]*) \z}xs) {
-    $result = $2 ne '' ? qr{(?$2)$1} : qr{$1};
-  } elsif ($subtest =~ m{^ m \s*  < (.+)  > ([a-z]*) \z}xs) {
-    $result = $2 ne '' ? qr{(?$2)$1} : qr{$1};
-  } elsif ($subtest =~ m{^ m \s* (\S) (.+) \1 ([a-z]*) \z}xs) {
-    $result = $2 ne '' ? qr{(?$2)$1} : qr{$1};
+  if ($subtest =~ m{^/ .+ / [a-z]* \z}xs ||
+      $subtest =~ m{^m (\W) .+ (\W) [a-z]* \z}xs) {
+    my ($rec, $err) = compile_regexp($subtest, 1);
+    if (!$rec) {
+      warn "askdns: subtest compile failed: '$subtest': $err\n";
+    } else {
+      $result = $rec;
+    }
   } elsif ($subtest =~ m{^ (["']) (.*) \1 \z}xs) {  # quoted string
     $result = $2;
   } elsif ($subtest =~ m{^ \[ ( (?:[A-Z]+|\d+)
@@ -306,7 +305,7 @@ sub set_config {
 
   push(@cmds, {
     setting => 'askdns',
-    is_admin => 1,
+    is_priv => 1,
     type => $Mail::SpamAssassin::Conf::CONF_TYPE_HASH_KEY_VALUE,
     code => sub {
       my($self, $key, $value, $line) = @_;
@@ -324,8 +323,9 @@ sub set_config {
         my @answer_types = split(/,/, $query_type);
         # https://www.iana.org/assignments/dns-parameters/dns-parameters.xml
         if (grep(!/^(?:ANY|A|AAAA|MX|TXT|PTR|NAPTR|NS|SOA|CERT|CNAME|DNAME|
-                       DHCID|HINFO|MINFO|RP|HIP|IPSECKEY|KX|LOC|SRV|
-                       SSHFP|SPF)\z/x, @answer_types)) {
+                       DHCID|HINFO|MINFO|RP|HIP|IPSECKEY|KX|LOC|GPOS|SRV|
+                       OPENPGPKEY|SSHFP|SPF|TLSA|URI|CAA|CSYNC)\z/x,
+                 @answer_types)) {
           return $Mail::SpamAssassin::Conf::INVALID_VALUE;
         }
         $query_type = 'ANY' if @answer_types > 1 || $answer_types[0] eq 'ANY';
@@ -333,25 +333,19 @@ sub set_config {
           $subtest = parse_and_canonicalize_subtest($subtest);
           defined $subtest or return $Mail::SpamAssassin::Conf::INVALID_VALUE;
         }
+
+        # initialize rule structure
+        $self->{askdns}{$rulename}{query} = $query_template;
+        $self->{askdns}{$rulename}{q_type} = $query_type;
+        $self->{askdns}{$rulename}{a_types} = \@answer_types;
+        $self->{askdns}{$rulename}{subtest} = $subtest;
+        $self->{askdns}{$rulename}{tags} = ();
+
         # collect tag names as used in each query template
-        my @tags = $query_template =~ /_([A-Z][A-Z0-9]*)_/g;
-        my %seen; @tags = grep(!$seen{$_}++, @tags);  # filter out duplicates
-
-        # group rules by tag names used in them (to be used as a hash key)
-        my $depends_on_tags = !@tags ? '' : join(',',@tags);
-
-        # subgroup rules by a DNS RR type and a nonexpanded query template
-        my $query_template_key = $query_type . ':' . $query_template;
-
-        $self->{askdns}{$depends_on_tags}{$query_template_key} ||=
-          { query => $query_template, rules => {}, q_type => $query_type,
-            a_types =>  # optimization: undef means "same as q_type"
-              @answer_types == 1 && $answer_types[0] eq $query_type ? undef
-                                                           : \@answer_types };
-        $self->{askdns}{$depends_on_tags}{$query_template_key}{rules}{$rulename}
-          = $subtest;
-      # dbg("askdns: rule: %s, config dep: %s, domkey: %s, subtest: %s",
-      #     $rulename, $depends_on_tags, $query_template_key, $subtest);
+        # also support common HEADER(arg) tag which does $pms->get(arg)
+        my @tags = $query_template =~ /_([A-Z][A-Z0-9]*(?:_[A-Z0-9]+)*(?:\(.*?\))?)_/g;
+        # save rule to tag dependencies
+        $self->{askdns}{$rulename}{tags}{$_} = 1 foreach (@tags);
 
         # just define the test so that scores and lint works
         $self->{parser}->add_test($rulename, undef,
@@ -366,182 +360,136 @@ sub set_config {
 # run as early as possible, launching DNS queries as soon as their
 # dependencies are fulfilled
 #
-sub parsed_metadata {
+sub check_dnsbl {
   my($self, $opts) = @_;
+
   my $pms = $opts->{permsgstatus};
   my $conf = $pms->{conf};
 
-  return if !$pms->is_dns_available;
-  $pms->{askdns_map_dnskey_to_rules} = {};
+  return if !$pms->is_dns_available();
 
   # walk through all collected askdns rules, obtain tag values whenever
   # they may become available, and launch DNS queries right after
-  #
-  for my $depends_on_tags (keys %{$conf->{askdns}}) {
-    my @tags;
-    @tags = split(/,/, $depends_on_tags)  if $depends_on_tags ne '';
-
-    if (would_log("dbg","askdns")) {
-      while ( my($query_template_key, $struct) =
-                each %{$conf->{askdns}{$depends_on_tags}} ) {
-        my($query_template, $query_type, $answer_types_ref, $rules) =
-          @$struct{qw(query q_type a_types rules)};
-        dbg("askdns: depend on tags %s, rules: %s ",
-            $depends_on_tags, join(', ', keys %$rules));
-      }
+  foreach my $rulename (keys %{$conf->{askdns}}) {
+    if (!$conf->{scores}->{$rulename}) {
+      dbg("askdns: skipping disabled rule $rulename");
+      next;
     }
-
-    if (!@tags) {
-      # no dependencies on tags, just call directly
-      $self->launch_queries($pms,$depends_on_tags);
-    } else {
-      # enqueue callback for tags needed
+    my @tags = sort keys %{$conf->{askdns}{$rulename}{tags}};
+    if (@tags) {
+      dbg("askdns: rule %s depends on tags: %s", $rulename,
+          join(', ', @tags));
       $pms->action_depends_on_tags(@tags == 1 ? $tags[0] : \@tags,
-              sub { my($pms,@args) = @_;
-                    $self->launch_queries($pms,$depends_on_tags) }
+            sub { my($pms,@args) = @_;
+                  $self->launch_queries($pms,$rulename,\@tags) }
       );
+    } else {
+      # no dependencies on tags, just call directly
+      $self->launch_queries($pms,$rulename,[]);
     }
   }
 }
 
-# generate DNS queries - called for each set of rules
-# when their tag dependencies are met
+# generate DNS queries - called for each rule when its tag dependencies
+# are met
 #
 sub launch_queries {
-  my($self, $pms, $depends_on_tags) = @_;
-  my $conf = $pms->{conf};
+  my($self, $pms, $rulename, $tags) = @_;
 
-  my %tags;
-  # obtain tag/value pairs of tags we depend upon in this set of rules
-  if ($depends_on_tags ne '') {
-    %tags = map( ($_,$pms->get_tag($_)), split(/,/,$depends_on_tags) );
+  my $arule = $pms->{conf}->{askdns}{$rulename};
+  my $query_tmpl = $arule->{query};
+  my $queries;
+  if (@$tags) {
+    if (!exists $pms->{askdns_qtmpl_cache}{$query_tmpl}) {
+      # replace tags in query template
+      # iterate through each tag, replacing list of strings as we go
+      my %q_iter = ( "$query_tmpl" => 1 );
+      foreach my $tag (@$tags) {
+        # cache tag values locally
+        if (!exists $pms->{askdns_tag_cache}{$tag}) {
+          my $valref = $pms->get_tag_raw($tag);
+          my @vals = grep { defined $_ && $_ ne '' } (ref $valref ? @$valref : $valref);
+          # Paranoid check for undefined tag
+          if (!@vals) {
+            dbg("askdns: skipping rule $rulename, no value found for tag: $tag");
+            return;
+          }
+          $pms->{askdns_tag_cache}{$tag} = \@vals;
+        }
+        my %q_iter_new;
+        foreach my $q (keys %q_iter) {
+          # handle space separated multi-valued tags
+          foreach my $val (@{$pms->{askdns_tag_cache}{$tag}}) {
+            if (exists $pms->{conf}->{uridnsbl_skip_domains}->{lc $val}) {
+              dbg("askdns: query skipped, uridnsbl_skip_domains: $val");
+              next;
+            }
+            my $dom = $pms->{main}->{registryboundaries}->trim_domain($val);
+            if (exists $pms->{conf}->{uridnsbl_skip_domains}->{lc $dom}) {
+              dbg("askdns: query skipped, uridnsbl_skip_domains: $val");
+              next;
+            }
+            my $qtmp = $q;
+            $qtmp =~ s/\Q_${tag}_\E/${val}/g;
+            $q_iter_new{$qtmp} = 1;
+          }
+        }
+        %q_iter = %q_iter_new;
+      }
+      # cache idn'd queries
+      my @q_arr;
+      push @q_arr, idn_to_ascii($_) foreach (keys %q_iter);
+      $pms->{askdns_qtmpl_cache}{$query_tmpl} = \@q_arr;
+    }
+    $queries = $pms->{askdns_qtmpl_cache}{$query_tmpl};
+  } else {
+    push @$queries, idn_to_ascii($query_tmpl);
   }
-  dbg("askdns: preparing queries which depend on tags: %s",
-      join(', ', map($_.' => '.$tags{$_}, keys %tags)));
 
-  # replace tag names in a query template with actual tag values
-  # and launch DNS queries
-  while ( my($query_template_key, $struct) =
-            each %{$conf->{askdns}{$depends_on_tags}} ) {
-    my($query_template, $query_type, $answer_types_ref, $rules) =
-      @$struct{qw(query q_type a_types rules)};
-
-    my @rulenames = keys %$rules;
-    if (grep($conf->{scores}->{$_}, @rulenames)) {
-      dbg("askdns: query template %s, type %s, rules: %s",
-          $query_template,
-          !$answer_types_ref ? $query_type
-            : $query_type.'/'.join(',',@$answer_types_ref),
-          join(', ', @rulenames));
-    } else {
-      dbg("askdns: query template %s, type %s, all rules disabled: %s",
-          $query_template, $query_type, join(', ', @rulenames));
+  foreach my $query (@$queries) {
+    if (!is_fqdn_valid($query, 1)) {
+      dbg("askdns: skipping invalid query ($rulename): $query");
       next;
     }
-
-    # collect all tag names from a template, each may occur more than once
-    my @templ_tags = $query_template =~ /_([A-Z][A-Z0-9]*)_/gs;
-
-    # filter out duplicate tag names, and tags with undefined or empty value
-    my %seen;
-    @templ_tags = grep(!$seen{$_}++ && defined $tags{$_} && $tags{$_} ne '',
-                       @templ_tags);
-
-    my %templ_vals;  # values that each tag takes
-    for my $t (@templ_tags) {
-      my %seen;
-      # a tag value may be a space-separated list,
-      # store it as an arrayref, removing duplicate values
-      $templ_vals{$t} = [ grep(!$seen{$_}++, split(' ',$tags{$t})) ];
-    }
-
-    # count through all tag value tuples
-    my @digit = (0) x @templ_tags;  # counting accumulator
-OUTER:
-    for (;;) {
-      my %current_tag_val;  # maps a tag name to its current iteration value
-      for my $j (0 .. $#templ_tags) {
-        my $t = $templ_tags[$j];
-        $current_tag_val{$t} = $templ_vals{$t}[$digit[$j]];
-      }
-      local $1;
-      my $query_domain = $query_template;
-      $query_domain =~ s{_([A-Z][A-Z0-9]*)_}
-                        { defined $current_tag_val{$1} ? $current_tag_val{$1}
-                                                       : '' }ge;
-
-      # the $dnskey identifies this query in AsyncLoop's pending_lookups
-      my $dnskey = join(':', 'askdns', $query_type, $query_domain);
-      dbg("askdns: expanded query %s, dns key %s", $query_domain, $dnskey);
-
-      if ($query_domain eq '') {
-        # ignore, just in case
-      } else {
-        if (!exists $pms->{askdns_map_dnskey_to_rules}{$dnskey}) {
-          $pms->{askdns_map_dnskey_to_rules}{$dnskey} =
-             [ [$query_type, $answer_types_ref, $rules] ];
-        } else {
-          push(@{$pms->{askdns_map_dnskey_to_rules}{$dnskey}},
-               [$query_type, $answer_types_ref, $rules] );
-        }
-        # launch a new DNS query for $query_type and $query_domain
-        my $ent = $pms->{async}->bgsend_and_start_lookup(
-          $query_domain, $query_type, undef,
-          { key => $dnskey, zone => $query_domain },
-          sub { my ($ent2,$pkt) = @_;
-                $self->process_response_packet($pms, $ent2, $pkt, $dnskey) },
-          master_deadline => $pms->{master_deadline} );
-        # these rules are now underway;  unless the rule hits, these will
-        # not be considered "finished" until harvest_dnsbl_queries() completes
-        $pms->register_async_rule_start($dnskey) if $ent;
-      }
-
-      last  if !@templ_tags;
-      # increment accumulator, little-endian
-      for (my $j = 0;  ; $j++) {
-        last  if ++$digit[$j] <= $#{$templ_vals{$templ_tags[$j]}};
-        $digit[$j] = 0;  # and carry
-        last OUTER  if $j >= $#templ_tags;
-      }
-    }
+    dbg("askdns: launching query ($rulename): $query");
+    my $ret = $pms->{async}->bgsend_and_start_lookup(
+      $query, $arule->{q_type}, undef,
+        { rulename => $rulename, type => 'AskDNS' },
+        sub { my ($ent,$pkt) = @_;
+              $self->process_response_packet($pms, $ent, $pkt, $rulename) },
+        master_deadline => $pms->{master_deadline}
+    );
+    $pms->rule_ready($rulename) if !$ret; # mark ready if nothing launched
   }
 }
 
 sub process_response_packet {
-  my($self, $pms, $ent, $pkt, $dnskey) = @_;
+  my($self, $pms, $ent, $pkt, $rulename) = @_;
 
-  my $conf = $pms->{conf};
-  my %rulenames_hit;
-
-  # map a dnskey back to info on queries which caused this DNS lookup
-  my $queries_ref = $pms->{askdns_map_dnskey_to_rules}{$dnskey};
-
-  my($header, @question, @answer, $qtype, $rcode);
   # NOTE: $pkt will be undef if the DNS query was aborted (e.g. timed out)
-  if ($pkt) {
-    @answer = $pkt->answer;
-    $header = $pkt->header;
-    @question = $pkt->question;
-    $qtype = uc $question[0]->qtype  if @question;
-    $rcode = uc $header->rcode  if $header;  # 'NOERROR', 'NXDOMAIN', ...
+  return if !$pkt;
 
-    # NOTE: qname is encoded in RFC 1035 zone format, decode it
-    dbg("askdns: answer received, rcode %s, query %s, answer has %d records",
-        $rcode,
-        join(', ', map(join('/', decode_dns_question_entry($_)), @question)),
-        scalar @answer);
+  my @question = $pkt->question;
+  return if !@question;
 
-    if (defined $rcode && exists $rcode_value{$rcode}) {
-      # Net::DNS return a rcode name for codes it knows about,
-      # and returns a number for the rest; we deal with numbers from here on
-      $rcode = $rcode_value{$rcode}  if exists $rcode_value{$rcode};
-    }
-  }
-  if (!@answer) {
-    # a trick to make the following loop run at least once, so that we can
-    # evaluate also rules which only care for rcode status
-    @answer = ( undef );
-  }
+  $pms->rule_ready($rulename); # mark rule ready for metas
+
+  my @answer = $pkt->answer;
+  my $rcode = uc $pkt->header->rcode;  # 'NOERROR', 'NXDOMAIN', ...
+
+  # NOTE: qname is encoded in RFC 1035 zone format, decode it
+  dbg("askdns: answer received (%s), rcode %s, query %s, answer has %d records",
+      $rulename, $rcode,
+      join(', ', map(join('/', decode_dns_question_entry($_)), @question)),
+      scalar @answer);
+
+  # Net::DNS return a rcode name for codes it knows about,
+  # and returns a number for the rest; we deal with numbers from here on
+  $rcode = $rcode_value{$rcode}  if exists $rcode_value{$rcode};
+
+  # a trick to make the following loop run at least once, so that we can
+  # evaluate also rules which only care for rcode status
+  @answer = (undef)  if !@answer;
 
   # NOTE:  $rr->rdstring returns the result encoded in a DNS zone file
   # format, i.e. enclosed in double quotes if a result contains whitespace
@@ -561,9 +509,12 @@ sub process_response_packet {
   # the code handling such reply from DNS MUST assemble all of these
   # marshaled text blocks into a single one before any syntactical
   # verification takes place.
-  # The same goes for RFC 4408 (SPF), RFC 4871 (DKIM), RFC 5617 (ADSP),
+  # The same goes for RFC 7208 (SPF), RFC 4871 (DKIM), RFC 5617 (ADSP),
   # draft-kucherawy-dmarc-base (DMARC), ...
 
+  my $arule = $pms->{conf}->{askdns}{$rulename};
+  my $subtest = $arule->{subtest};
+
   for my $rr (@answer) {
     my($rr_rdatastr, $rdatanum, $rr_type);
     if (!$rr) {
@@ -571,72 +522,62 @@ sub process_response_packet {
     } else {
       $rr_type = uc $rr->type;
       if ($rr_type eq 'A') {
-        # Net::DNS::RR::A::address() is available since Net::DNS 0.69
-        $rr_rdatastr = $rr->UNIVERSAL::can('address') ? $rr->address
-                                                      : $rr->rdatastr;
+        $rr_rdatastr = $rr->address;
         if ($rr_rdatastr =~ m/^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\z/) {
           $rdatanum = Mail::SpamAssassin::Util::my_inet_aton($rr_rdatastr);
         }
 
       } elsif ($rr->UNIVERSAL::can('txtdata')) {
         # TXT, SPF: join with no intervening spaces, as per RFC 5518
-        if ($txtdata_can_provide_a_list || $rr_type ne 'TXT') {
-          $rr_rdatastr = join('', $rr->txtdata);  # txtdata() in list context!
-        } else {  # char_str_list() is only available for TXT records
-          $rr_rdatastr = join('', $rr->char_str_list);  # historical
-        }
+        $rr_rdatastr = join('', $rr->txtdata);  # txtdata() in list context!
+        # Net::DNS attempts to decode text strings in a TXT record as UTF-8,
+        # which is undesired: octets failing the UTF-8 decoding are converted
+        # to a Unicode "replacement character" U+FFFD (encoded as octets
+        # \x{EF}\x{BF}\x{BD} in UTF-8), and ASCII text is unnecessarily
+        # flagged as perl native characters (utf8 flag on), which can be
+        # disruptive on later processing, e.g. implicitly upgrading strings
+        # on concatenation. Unfortunately there is no way of legally bypassing
+        # the UTF-8 decoding by Net::DNS::RR::TXT in Net::DNS::RR::Text.
+        # Try to minimize damage by encoding back to UTF-8 octets:
+        utf8::encode($rr_rdatastr)  if utf8::is_utf8($rr_rdatastr);
+
       } else {
-        # rdatastr() is historical, use rdstring() since Net::DNS 0.69
-        $rr_rdatastr = $rr->UNIVERSAL::can('rdstring') ? $rr->rdstring
-                                                       : $rr->rdatastr;
+        $rr_rdatastr = $rr->rdstring;
         utf8::encode($rr_rdatastr)  if utf8::is_utf8($rr_rdatastr);
       }
-    # dbg("askdns: received rr type %s, data: %s", $rr_type, $rr_rdatastr);
+      # dbg("askdns: received rr type %s, data: %s", $rr_type, $rr_rdatastr);
     }
 
-    for my $q_tuple (!ref $queries_ref ? () : @$queries_ref) {
-      next  if !$q_tuple;
-      my($query_type, $answer_types_ref, $rules) = @$q_tuple;
-
-      next  if !defined $qtype;
-      $answer_types_ref = [$query_type]  if !defined $answer_types_ref;
-
-      while (my($rulename,$subtest) = each %$rules) {
-        my $match;
-        local($1,$2,$3);
-        if (ref $subtest eq 'HASH') {  # a list of DNS rcodes (as hash keys)
-          $match = 1  if $subtest->{$rcode};
-        } elsif ($rcode != 0) {
-          # skip remaining tests on DNS error
-        } elsif (!defined($rr_type) ||
-                 !grep($_ eq 'ANY' || $_ eq $rr_type, @$answer_types_ref) ) {
-          # skip remaining tests on wrong RR type
-        } elsif (!defined $subtest) {
-          $match = 1;  # any valid response of the requested RR type matches
-        } elsif (ref $subtest eq 'Regexp') {  # a regular expression
-          $match = 1  if $rr_rdatastr =~ $subtest;
-        } elsif ($rr_rdatastr eq $subtest) {  # exact equality
-          $match = 1;
-        } elsif (defined $rdatanum &&
-                 $subtest =~ m{^ (\d+) (?: ([/-]) (\d+) )? \z}x) {
-          my($n1,$delim,$n2) = ($1,$2,$3);
-          $match =
-            !defined $n2  ? ($rdatanum & $n1) &&                  # mask only
-                              (($rdatanum & 0xff000000) == 0x7f000000)  # 127/8
-          : $delim eq '-' ? $rdatanum >= $n1 && $rdatanum <= $n2  # range
-          : $delim eq '/' ? ($rdatanum & $n2) == (int($n1) & $n2) # value/mask
-          : 0; # notice int($n1) to fix perl ~5.14 taint bug (Bug 7725)
-        }
-        if ($match) {
-          $self->askdns_hit($pms, $ent->{query_domain}, $qtype,
-                            $rr_rdatastr, $rulename);
-          $rulenames_hit{$rulename} = 1;
-        }
-      }
+    my $match;
+    local($1,$2,$3);
+    if (ref $subtest eq 'HASH') {  # a list of DNS rcodes (as hash keys)
+      $match = 1  if $subtest->{$rcode};
+    } elsif ($rcode != 0) {
+      # skip remaining tests on DNS error
+    } elsif (!defined($rr_type) ||
+             !grep($_ eq 'ANY' || $_ eq $rr_type, @{$arule->{a_types}}) ) {
+      # skip remaining tests on wrong RR type
+    } elsif (!defined $subtest) {
+      $match = 1;  # any valid response of the requested RR type matches
+    } elsif (ref $subtest eq 'Regexp') {  # a regular expression
+      $match = 1  if $rr_rdatastr =~ $subtest;
+    } elsif ($rr_rdatastr eq $subtest) {  # exact equality
+      $match = 1;
+    } elsif (defined $rdatanum &&
+             $subtest =~ m{^ (\d+) (?: ([/-]) (\d+) )? \z}x) {
+      my($n1,$delim,$n2) = ($1,$2,$3);
+      $match =
+        !defined $n2 ? ($rdatanum & $n1) &&                     # mask only
+                       (($rdatanum & 0xff000000) == 0x7f000000) # 127/8
+        : $delim eq '-' ? $rdatanum >= $n1 && $rdatanum <= $n2  # range
+        : $delim eq '/' ? ($rdatanum & $n2) == (int($n1) & $n2) # value/mask
+        : 0; # notice int($n1) to fix perl ~5.14 taint bug (Bug 7725)
+    }
+    if ($match) {
+      $self->askdns_hit($pms, $ent->{query_domain}, $question[0]->qtype,
+                        $rr_rdatastr, $rulename);
     }
   }
-  # these rules have completed (since they got at least 1 hit)
-  $pms->register_async_rule_finish($_)  for keys %rulenames_hit;
 }
 
 sub askdns_hit {
@@ -648,9 +589,11 @@ sub askdns_hit {
 
   # only the first hit will show in the test log report, even if
   # an answer section matches more than once - got_hit() handles this
-  $pms->clear_test_state;
-  $pms->test_log(sprintf("%s %s:%s", $query_domain,$qtype,$rr_rdatastr));
+  $pms->test_log(sprintf("%s %s:%s", $query_domain,$qtype,$rr_rdatastr), $rulename);
   $pms->got_hit($rulename, 'ASKDNS: ', ruletype => 'askdns');  # score=>$score
 }
 
+# Version features
+sub has_tag_header { 1 } # HEADER() was implemented together with Conf::feature_get_host # Bug 7734
+
 1;
