@@ -38,6 +38,9 @@ DNSEVAL - look up URLs against DNS blocklists
 The DNSEval plugin queries dns to see if a domain or an ip address
 present on one of email's headers is on a particular rbl.
 
+Any host or its domain part matching uridnsbl_skip_domains is ignored
+by default.
+
 =cut
 
 package Mail::SpamAssassin::Plugin::DNSEval;
@@ -45,7 +48,7 @@ package Mail::SpamAssassin::Plugin::DNSEval;
 use Mail::SpamAssassin::Plugin;
 use Mail::SpamAssassin::Logger;
 use Mail::SpamAssassin::Constants qw(:ip);
-use Mail::SpamAssassin::Util qw(reverse_ip_address is_fqdn_valid);
+use Mail::SpamAssassin::Util qw(reverse_ip_address idn_to_ascii compile_regexp is_fqdn_valid);
 
 use strict;
 use warnings;
@@ -55,7 +58,6 @@ use re 'taint';
 our @ISA = qw(Mail::SpamAssassin::Plugin);
 
 my $IP_ADDRESS = IP_ADDRESS;
-my $IP_PRIVATE = IP_PRIVATE;
 
 # constructor: register the eval rule
 sub new {
@@ -75,7 +77,6 @@ sub new {
     'check_rbl_ns_from',
     'check_rbl_txt',
     'check_rbl_sub',
-    'check_rbl_results_for',
     'check_rbl_from_host',
     'check_rbl_from_domain',
     'check_rbl_envfrom',
@@ -86,7 +87,7 @@ sub new {
 
   $self->set_config($mailsaobject->{conf});
   foreach(@{$self->{'evalrules'}}) {
-    $self->register_eval_rule($_);
+    $self->register_eval_rule($_, $Mail::SpamAssassin::Conf::TYPE_RBL_EVALS);
   }
 
   return $self;
@@ -143,6 +144,7 @@ sub set_config {
     my @cmds;
     push(@cmds, {
         setting => 'rbl_headers',
+        is_priv => 1,
         default => 'EnvelopeFrom,Reply-To,Disposition-Notification-To,X-WebmailclientIP,X-Source-IP',
         type => $Mail::SpamAssassin::Conf::CONF_TYPE_STRING,
         }
@@ -154,21 +156,104 @@ sub set_config {
 # directly as part of PMS
 sub check_start {
   my ($self, $opts) = @_;
+  my $pms = $opts->{permsgstatus};
 
   foreach(@{$self->{'evalrules'}}) {
-    $opts->{'permsgstatus'}->register_plugin_eval_glue($_);
+    $pms->register_plugin_eval_glue($_);
   }
+
+  # Initialize check_rbl_sub tests
+  $self->_init_rbl_subs($pms);
+}
+
+sub _init_rbl_subs {
+  my ($self, $pms) = @_;
+  my $conf = $pms->{conf};
+
+  # Very hacky stuff and direct rbl_evals usage for now, TODO rewrite everything
+  foreach my $rule (@{$conf->{eval_to_rule}->{check_rbl_sub}||[]}) {
+    next if !exists $conf->{rbl_evals}->{$rule};
+    next if !$conf->{scores}->{$rule};
+    # rbl_evals is [$function,[@args]]
+    my $args = $conf->{rbl_evals}->{$rule}->[1];
+    my ($set, $subtest) = @$args;
+    if (!defined $subtest) {
+      warn("dnseval: missing subtest for rule $rule\n");
+      next;
+    }
+    if ($subtest =~ /^sb:/) {
+      warn("dnseval: ignored $rule, SenderBase rules are deprecated\n");
+      next;
+    }
+    # Compile as regex if not pure ip/bitmask (same check in process_dnsbl_result)
+    if ($subtest !~ /^\d+(?:\.\d+\.\d+\.\d+)?$/) {
+      my ($rec, $err) = compile_regexp($subtest, 0);
+      if (!$rec) {
+        warn("dnseval: invalid rule $rule subtest regexp '$subtest': $err\n");
+        next;
+      }
+      $subtest = $rec;
+    }
+    dbg("dnseval: initialize check_rbl_sub for rule $rule, set $set, subtest $subtest");
+    push @{$pms->{rbl_subs}{$set}}, [$subtest, $rule];
+  }
+}
+
+sub parsed_metadata {
+  my ($self, $opts) = @_;
+
+  my $pms = $opts->{permsgstatus};
+
+  return 1 if $self->{main}->{conf}->{skip_rbl_checks};
+  return 1 if !$pms->is_dns_available();
+
+  # Process relaylists only once, not everytime in check_rbl_backend
+  #
+  # ok, make a list of all the IPs in the untrusted set
+  my @fullips = map { $_->{ip} } @{$pms->{relays_untrusted}};
+  # now, make a list of all the IPs in the external set, for use in
+  # notfirsthop testing.  This will often be more IPs than found
+  # in @fullips.  It includes the IPs that are trusted, but
+  # not in internal_networks.
+  my @fullexternal = map {
+	(!$_->{internal}) ? ($_->{ip}) : ()
+      } @{$pms->{relays_trusted}};
+  push @fullexternal, @fullips; # add untrusted set too
+  # Make sure a header significantly improves results before adding here
+  # X-Sender-Ip: could be worth using (very low occurence for me)
+  # X-Sender: has a very low bang-for-buck for me
+  my @originating;
+  foreach my $header (@{$pms->{conf}->{originating_ip_headers}}) {
+    my $str = $pms->get($header, undef);
+    next unless defined $str && $str ne '';
+    push @originating, ($str =~ m/($IP_ADDRESS)/g);
+  }
+  # Let's go ahead and trim away all private ips (KLC)
+  # also uniq the list and strip dups. (jm)
+  my @ips = $self->ip_list_uniq_and_strip_private(@fullips);
+  # if there's no untrusted IPs, it means we trust all the open-internet
+  # relays, so we skip checks
+  if (scalar @ips + scalar @originating > 0) {
+    dbg("dnseval: IPs found: full-external: ".join(", ", @fullexternal).
+      " untrusted: ".join(", ", @ips).
+      " originating: ".join(", ", @originating));
+    @{$pms->{dnseval_fullexternal}} = @fullexternal;
+    @{$pms->{dnseval_ips}} = @ips;
+    @{$pms->{dnseval_originating}} = @originating;
+  }
+
+  return 1;
 }
 
 sub ip_list_uniq_and_strip_private {
   my ($self, @origips) = @_;
   my @ips;
   my %seen;
-  my $IP_PRIVATE = IP_PRIVATE;
   foreach my $ip (@origips) {
     next unless $ip;
-    next if (exists ($seen{$ip})); $seen{$ip} = 1;
-    next if ($ip =~ /$IP_PRIVATE/o);
+    next if exists $seen{$ip};
+    $seen{$ip} = 1;
+    next if $ip =~ IS_IP_PRIVATE;
     push(@ips, $ip);
   }
   return @ips;
@@ -181,12 +266,17 @@ sub ip_list_uniq_and_strip_private {
 sub check_rbl_accreditor {
   my ($self, $pms, $rule, $set, $rbl_server, $subtest, $accreditor) = @_;
 
+  return 0 if $self->{main}->{conf}->{skip_rbl_checks};
+  return 0 if !$pms->is_dns_available();
+
   if (!defined $pms->{accreditor_tag}) {
     $self->message_accreditor_tag($pms);
   }
   if ($pms->{accreditor_tag}->{$accreditor}) {
-    $self->check_rbl_backend($pms, $rule, $set, $rbl_server, 'A', $subtest);
+    # return undef for async status
+    return $self->_check_rbl_backend($pms, $rule, $set, $rbl_server, 'A', $subtest);
   }
+
   return 0;
 }
 
@@ -226,58 +316,16 @@ sub message_accreditor_tag {
   $pms->{accreditor_tag} = \%acctags;
 }
 
-sub check_rbl_backend {
+sub _check_rbl_backend {
   my ($self, $pms, $rule, $set, $rbl_server, $type, $subtest) = @_;
-  local ($_);
 
-  # First check that DNS is available, if not do not perform this check
-  return 0 if $self->{main}->{conf}->{skip_rbl_checks};
-  return 0 unless $pms->is_dns_available();
+  return if !exists $pms->{dnseval_ips}; # no untrusted ips
 
-  if (($rbl_server !~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/) &&
-      (index($rbl_server, '.') >= 0) &&
-      ($rbl_server !~ /\.$/)) {
-    $rbl_server .= ".";
-  }
-
-  dbg("dns: checking RBL $rbl_server, set $set");
-
-  # ok, make a list of all the IPs in the untrusted set
-  my @fullips = map { $_->{ip} } @{$pms->{relays_untrusted}};
-
-  # now, make a list of all the IPs in the external set, for use in
-  # notfirsthop testing.  This will often be more IPs than found
-  # in @fullips.  It includes the IPs that are trusted, but
-  # not in internal_networks.
-  my @fullexternal = map {
-	(!$_->{internal}) ? ($_->{ip}) : ()
-      } @{$pms->{relays_trusted}};
-  push (@fullexternal, @fullips);	# add untrusted set too
-
-  # Make sure a header significantly improves results before adding here
-  # X-Sender-Ip: could be worth using (very low occurence for me)
-  # X-Sender: has a very low bang-for-buck for me
-  my $IP_ADDRESS = IP_ADDRESS;
-  my @originating;
-  for my $header (@{$pms->{conf}->{originating_ip_headers}}) {
-    my $str = $pms->get($header,undef);
-    next unless defined $str && $str ne '';
-    push (@originating, ($str =~ m/($IP_ADDRESS)/g));
-  }
-
-  # Let's go ahead and trim away all private ips (KLC)
-  # also uniq the list and strip dups. (jm)
-  my @ips = $self->ip_list_uniq_and_strip_private(@fullips);
-
-  # if there's no untrusted IPs, it means we trust all the open-internet
-  # relays, so we can return right now.
-  return 0 unless (scalar @ips + scalar @originating > 0);
-
-  dbg("dns: IPs found: full-external: ".join(", ", @fullexternal).
-	" untrusted: ".join(", ", @ips).
-	" originating: ".join(", ", @originating));
+  $rbl_server =~ s/\.+\z//; # strip unneeded trailing dot
+  dbg("dnseval: checking RBL $rbl_server, set $set, rule $rule");
 
   my $trusted = $self->{main}->{conf}->{trusted_networks};
+  my @ips = @{$pms->{dnseval_ips}};
 
   # If name is foo-notfirsthop, check all addresses except for
   # the originating one.  Suitable for use with dialup lists, like the PDL.
@@ -293,9 +341,9 @@ sub check_rbl_backend {
     # specified some third-party relays as trusted.  Also, don't use
     # @originating; those headers are added by a phase of relaying through
     # a server like Hotmail, which is not going to be in dialup lists anyway.
-    @ips = $self->ip_list_uniq_and_strip_private(@fullexternal);
+    @ips = $self->ip_list_uniq_and_strip_private(@{$pms->{dnseval_fullexternal}});
     if ($1 eq "lastexternal") {
-      @ips = (defined $ips[0]) ? ($ips[0]) : ();
+      @ips = defined $ips[0] ? ($ips[0]) : ();
     } else {
 	pop @ips if (scalar @ips > 1);
     }
@@ -307,14 +355,14 @@ sub check_rbl_backend {
   elsif ($set =~ /-(first|un)trusted$/)
   {
     my @tips;
-    foreach my $ip (@originating) {
+    foreach my $ip (@{$pms->{dnseval_originating}}) {
       if ($ip && !$trusted->contains_ip($ip)) {
         push(@tips, $ip);
       }
     }
-    @ips = $self->ip_list_uniq_and_strip_private (@ips, @tips);
+    @ips = $self->ip_list_uniq_and_strip_private(@ips, @tips);
     if ($1 eq "first") {
-      @ips = (defined $ips[0]) ? ($ips[0]) : ();
+      @ips = defined $ips[0] ? ($ips[0]) : ();
     } else {
       shift @ips;
     }
@@ -322,7 +370,7 @@ sub check_rbl_backend {
   else
   {
     my @tips;
-    foreach my $ip (@originating) {
+    foreach my $ip (@{$pms->{dnseval_originating}}) {
       if ($ip && !$trusted->contains_ip($ip)) {
         push(@tips, $ip);
       }
@@ -333,71 +381,77 @@ sub check_rbl_backend {
   }
 
   # How many IPs max you check in the received lines
-  my $checklast=$self->{main}->{conf}->{num_check_received};
+  my $checklast = $self->{main}->{conf}->{num_check_received};
 
   if (scalar @ips > $checklast) {
     splice (@ips, $checklast);	# remove all others
   }
 
-  my $tflags = $pms->{conf}->{tflags}->{$rule};
-
   # Trusted relays should only be checked against nice rules (dnswls)
-  if (defined $tflags && $tflags !~ /\bnice\b/) {
+  if (($pms->{conf}->{tflags}->{$rule}||'') !~ /\bnice\b/) {
     # remove trusted hosts from beginning
     while (@ips && $trusted->contains_ip($ips[0])) { shift @ips }
   }
 
   unless (scalar @ips > 0) {
-    dbg("dns: no untrusted IPs to check");
+    dbg("dnseval: no untrusted IPs to check");
     return 0;
   }
 
-  dbg("dns: only inspecting the following IPs: ".join(", ", @ips));
+  dbg("dnseval: only inspecting the following IPs: ".join(", ", @ips));
 
-  eval {
-    foreach my $ip (@ips) {
-      my $revip = reverse_ip_address($ip);
-      $pms->do_rbl_lookup($rule, $set, $type,
-                          $revip.'.'.$rbl_server, $subtest) if defined $revip;
+  my $queries;
+  foreach my $ip (@ips) {
+    if (defined(my $revip = reverse_ip_address($ip))) {
+      my $ret = $pms->do_rbl_lookup($rule, $set, $type, $revip.'.'.$rbl_server, $subtest);
+      $queries++ if defined $ret;
     }
-  };
+  }
 
   # note that results are not handled here, hits are handled directly
   # as DNS responses are harvested
-  return 0;
+  return 0 if !$queries; # no query started
+  return; # return undef for async status
 }
 
 sub check_rbl {
   my ($self, $pms, $rule, $set, $rbl_server, $subtest) = @_;
-  $self->check_rbl_backend($pms, $rule, $set, $rbl_server, 'A', $subtest);
+
+  return 0 if $self->{main}->{conf}->{skip_rbl_checks};
+  return 0 if !$pms->is_dns_available();
+
+  # return undef for async status
+  return $self->_check_rbl_backend($pms, $rule, $set, $rbl_server, 'A', $subtest);
 }
 
 sub check_rbl_txt {
   my ($self, $pms, $rule, $set, $rbl_server, $subtest) = @_;
-  $self->check_rbl_backend($pms, $rule, $set, $rbl_server, 'TXT', $subtest);
-}
-
-# run for first message 
-sub check_rbl_sub {
-  my ($self, $pms, $rule, $set, $subtest) = @_;
 
   return 0 if $self->{main}->{conf}->{skip_rbl_checks};
-  return 0 unless $pms->is_dns_available();
+  return 0 if !$pms->is_dns_available();
 
-  $pms->register_rbl_subtest($rule, $set, $subtest);
+  # return undef for async status
+  return $self->_check_rbl_backend($pms, $rule, $set, $rbl_server, 'TXT', $subtest);
 }
 
-# backward compatibility
-sub check_rbl_results_for {
-  #warn "dns: check_rbl_results_for() is deprecated, use check_rbl_sub()\n";
-  check_rbl_sub(@_);
+sub check_rbl_sub {
+  my ($self, $pms, $rule, $set, $subtest) = @_;
+  # just a dummy, _init_rbl_subs/do_rbl_lookup handles the subs
+
+  return; # return undef for async status
 }
 
 # this only checks the address host name and not the domain name because
 # using the domain name had much worse results for dsn.rfc-ignorant.org
 sub check_rbl_from_host {
   my ($self, $pms, $rule, $set, $rbl_server, $subtest) = @_; 
-  _check_rbl_addresses($self, $pms, $rule, $set, $rbl_server, $subtest, $pms->all_from_addrs());
+
+  return 0 if $self->{main}->{conf}->{skip_rbl_checks};
+  return 0 if !$pms->is_dns_available();
+
+  # return undef for async status
+  return $self->_check_rbl_addresses($pms, $rule, $set, $rbl_server,
+                                     $subtest, $pms->all_from_addrs());
 }
 
 sub check_rbl_headers {
@@ -415,16 +469,18 @@ sub check_rbl_headers {
     @env_hdr = split(/,/, $conf->{rbl_headers});
   }
 
+  my $queries;
   foreach my $rbl_headers (@env_hdr) {
     my $addr = $pms->get($rbl_headers.':addr', undef);
     if ( defined $addr && $addr =~ /\@([^\@\s]+)/ ) {
-      $self->_check_rbl_addresses($pms, $rule, $set, $rbl_server,
-        $subtest, $addr);
+      my $ret = $self->_check_rbl_addresses($pms, $rule, $set, $rbl_server,
+                                            $subtest, $addr);
+      $queries++ if defined $ret;
     } else {
       my $unsplitted_host = $pms->get($rbl_headers);
       chomp($unsplitted_host);
       foreach my $host (split(/\n/, $unsplitted_host)) {
-        if($host =~ /^$IP_ADDRESS$/ ) {
+        if ($host =~ IS_IP_ADDRESS) {
           next if ($conf->{tflags}->{$rule}||'') =~ /\bdomains_only\b/;
           $host = reverse_ip_address($host);
         } else {
@@ -432,17 +488,23 @@ sub check_rbl_headers {
           next unless is_fqdn_valid($host);
           next unless $pms->{main}->{registryboundaries}->is_domain_valid($host);
         }
-        $pms->do_rbl_lookup($rule, $set, 'A', "$host.$rbl_server", $subtest);
+        my $ret = $pms->do_rbl_lookup($rule, $set, 'A', "$host.$rbl_server", $subtest);
+        $queries++ if defined $ret;
       }
     }
   }
+
+  return 0 if !$queries; # no query started
+  return; # return undef for async status
 }
 
 =over 4
 
 =item check_rbl_from_domain
 
-This checks all the from addrs domain names as an alternate to check_rbl_from_host.  As of v3.4.1, it has been improved to include a subtest for a specific octet.
+This checks all the from addrs domain names as an alternate to
+check_rbl_from_host.  As of v3.4.1, it has been improved to include a
+subtest for a specific octet.
 
 =back
 
@@ -450,7 +512,13 @@ This checks all the from addrs domain names as an alternate to check_rbl_from_ho
 
 sub check_rbl_from_domain {
   my ($self, $pms, $rule, $set, $rbl_server, $subtest) = @_;
-  _check_rbl_addresses($self, $pms, $rule, $set, $rbl_server, $subtest, $pms->all_from_addrs_domains());
+
+  return 0 if $self->{main}->{conf}->{skip_rbl_checks};
+  return 0 if !$pms->is_dns_available();
+
+  # return undef for async status
+  return $self->_check_rbl_addresses($pms, $rule, $set, $rbl_server,
+                                     $subtest, $pms->all_from_addrs_domains());
 }
 
 =over 4
@@ -472,6 +540,7 @@ sub check_rbl_ns_from {
   return 0 if $self->{main}->{conf}->{skip_rbl_checks};
   return 0 unless $pms->is_dns_available();
 
+  dbg("dnseval: EnvelopeFrom header not found") unless defined (($pms->get("EnvelopeFrom:addr"))[0]);
   for my $from ($pms->get('EnvelopeFrom:addr')) {
     next unless defined $from;
     $from =~ tr/././s;          # bug 3366
@@ -482,20 +551,29 @@ sub check_rbl_ns_from {
   }
   return 0 unless defined $domain;
 
-  dbg("dns: checking NS for host $domain");
+  dbg("dnseval: checking NS for host $domain");
 
-  my $key = "NS:" . $domain;
   my $obj = { dom => $domain, rule => $rule, set => $set, rbl_server => $rbl_server, subtest => $subtest };
   my $ent = {
-    rulename => $rule, key => $key, zone => $domain, obj => $obj, type => "URI-NS",
+    rulename => $rule, zone => $domain, obj => $obj, type => "URI-NS",
   };
+  if (exists $pms->{conf}->{uridnsbl_skip_domains}->{lc $domain}) {
+    dbg("dnseval: query skipped, uridnsbl_skip_domains: $domain");
+    return 0;
+  }
+  my $dom = $pms->{main}->{registryboundaries}->trim_domain($domain);
+  if (exists $pms->{conf}->{uridnsbl_skip_domains}->{lc $dom}) {
+    dbg("dnseval: query skipped, uridnsbl_skip_domains: $domain");
+    return 0;
+  }
   # dig $dom ns
-  $ent = $pms->{async}->bgsend_and_start_lookup(
+  my $ret = $pms->{async}->bgsend_and_start_lookup(
     $domain, 'NS', undef, $ent,
     sub { my ($ent2,$pkt) = @_;
           $self->complete_ns_lookup($pms, $ent2, $pkt, $domain) },
     master_deadline => $pms->{master_deadline} );
-  return $ent;
+  return 0 if !defined $ret; # no query started
+  return; # return undef for async status
 }
 
 sub complete_ns_lookup {
@@ -508,11 +586,11 @@ sub complete_ns_lookup {
 
   if (!$pkt) {
     # $pkt will be undef if the DNS query was aborted (e.g. timed out)
-    dbg("DNSEval: complete_ns_lookup aborted %s", $ent->{key});
+    dbg("dnseval: complete_ns_lookup aborted %s", $ent->{key});
     return;
   }
 
-  dbg("DNSEval: complete_ns_lookup %s", $ent->{key});
+  dbg("dnseval: complete_ns_lookup %s", $ent->{key});
   my @ns = $pkt->authority;
 
   foreach my $rr (@ns) {
@@ -521,9 +599,9 @@ sub complete_ns_lookup {
     chomp($nshost);
     if (is_fqdn_valid($nshost)) {
       if ( defined $subtest ) {
-        dbg("dns: checking [$nshost] / $rule / $set / $rbl_server / $subtest");
+        dbg("dnseval: checking [$nshost] / $rule / $set / $rbl_server / $subtest");
       } else {
-        dbg("dns: checking [$nshost] / $rule / $set / $rbl_server");
+        dbg("dnseval: checking [$nshost] / $rule / $set / $rbl_server");
       }
       $pms->do_rbl_lookup($rule, $set, 'A',
         "$nshost.$rbl_server", $subtest);
@@ -548,7 +626,7 @@ sub check_rbl_rcvd {
   my @udnsrcvd = ();
 
   return 0 if $self->{main}->{conf}->{skip_rbl_checks};
-  return 0 if !$pms->is_dns_available();  
+  return 0 if !$pms->is_dns_available();
 
   my $rcvd = $pms->{relays_untrusted}->[$pms->{num_relays_untrusted} - 1];
   my @dnsrcvd = ( $rcvd->{ip}, $rcvd->{by}, $rcvd->{helo}, $rcvd->{rdns} );
@@ -559,10 +637,11 @@ sub check_rbl_rcvd {
     }
   }
 
+  my $queries;
   foreach my $host ( @udnsrcvd ) {
     if((defined $host) and ($host ne "")) {
       chomp($host);
-      if($host =~ /^$IP_ADDRESS$/ ) {
+      if ($host =~ IS_IP_ADDRESS) {
         next if ($pms->{conf}->{tflags}->{$rule}||'') =~ /\bdomains_only\b/;
         $host = reverse_ip_address($host);
       } else {
@@ -572,73 +651,91 @@ sub check_rbl_rcvd {
         next unless $pms->{main}->{registryboundaries}->is_domain_valid($host);
       }
       if ( defined $subtest ) {
-        dbg("dns: checking [$host] / $rule / $set / $rbl_server / $subtest");
+        dbg("dnseval: checking [$host] / $rule / $set / $rbl_server / $subtest");
       } else {
-        dbg("dns: checking [$host] / $rule / $set / $rbl_server");
+        dbg("dnseval: checking [$host] / $rule / $set / $rbl_server");
       }
-      $pms->do_rbl_lookup($rule, $set, 'A', "$host.$rbl_server", $subtest);
+      my $ret = $pms->do_rbl_lookup($rule, $set, 'A', "$host.$rbl_server", $subtest);
+      $queries++ if defined $ret;
     }
   }
-  return 0;
+
+  return 0 if !$queries; # no query started
+  return; # return undef for async status
 }
 
 # this only checks the address host name and not the domain name because
 # using the domain name had much worse results for dsn.rfc-ignorant.org
 sub check_rbl_envfrom {
   my ($self, $pms, $rule, $set, $rbl_server, $subtest) = @_; 
-  _check_rbl_addresses($self, $pms, $rule, $set, $rbl_server, $subtest, $pms->get('EnvelopeFrom:addr',undef));
+
+  return 0 if $self->{main}->{conf}->{skip_rbl_checks};
+  return 0 if !$pms->is_dns_available();
+
+  # return undef for async status
+  return $self->_check_rbl_addresses($pms, $rule, $set, $rbl_server,
+                                 $subtest, $pms->get('EnvelopeFrom:addr',undef));
 }
 
 sub _check_rbl_addresses {
   my ($self, $pms, $rule, $set, $rbl_server, $subtest, @addresses) = @_;
   
-  return 0 if $self->{main}->{conf}->{skip_rbl_checks};
-  return 0 unless $pms->is_dns_available();
+  $rbl_server =~ s/\.+\z//; # strip unneeded trailing dot
 
   my %hosts;
   for (@addresses) {
-    next if !defined($_) || !/ \@ ( [^\@\s]+ )/x;
+    next if !defined($_) || !/\@([^\@\s]+)/;
     my $address = $1;
     # strip leading & trailing dots (as seen in some e-mail addresses)
-    $address =~ s/^\.+//; $address =~ s/\.+\z//;
+    $address =~ s/^\.+//;
+    $address =~ s/\.+\z//;
     # squash duplicate dots to avoid an invalid DNS query with a null label
-    $address =~ tr/.//s;
-    $hosts{lc($address)} = 1  if $address =~ /\./;  # must by a FQDN
+    # Also checks it's FQDN
+    if ($address =~ tr/.//s) {
+      $hosts{lc($address)} = 1;
+    }
   }
   return unless scalar keys %hosts;
 
-  if (($rbl_server !~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/) &&
-      (index($rbl_server, '.') >= 0) &&
-      ($rbl_server !~ /\.$/)) {
-    $rbl_server .= ".";
-  }
-
+  my $queries;
   for my $host (keys %hosts) {
-    if ($host =~ /^$IP_ADDRESS$/) {
+    if ($host =~ IS_IP_ADDRESS) {
       next if ($pms->{conf}->{tflags}->{$rule}||'') =~ /\bdomains_only\b/;
       $host = reverse_ip_address($host);
     } else {
       next if ($pms->{conf}->{tflags}->{$rule}||'') =~ /\bips_only\b/;
       next unless is_fqdn_valid($host);
       next unless $pms->{main}->{registryboundaries}->is_domain_valid($host);
+      if (exists $pms->{conf}->{uridnsbl_skip_domains}->{lc $host}) {
+        dbg("dnseval: query skipped, uridnsbl_skip_domains: $host");
+        next;
+      }
+      my $dom = $pms->{main}->{registryboundaries}->trim_domain($host);
+      if (exists $pms->{conf}->{uridnsbl_skip_domains}->{lc $dom}) {
+        dbg("dnseval: query skipped, uridnsbl_skip_domains: $host");
+        next;
+      }
     }
-    dbg("dns: checking [$host] / $rule / $set / $rbl_server");
-    $pms->do_rbl_lookup($rule, $set, 'A', "$host.$rbl_server", $subtest);
+    dbg("dnseval: checking [$host] / $rule / $set / $rbl_server");
+    my $ret = $pms->do_rbl_lookup($rule, $set, 'A', "$host.$rbl_server", $subtest);
+    $queries++ if defined $ret;
   }
+
+  return 0 if !$queries; # no async
+  return; # return undef for async status
 }
 
 sub check_dns_sender {
   my ($self, $pms, $rule) = @_;
 
   return 0 if $self->{main}->{conf}->{skip_rbl_checks};
-  return 0 unless $pms->is_dns_available();
+  return 0 if !$pms->is_dns_available();
 
   my $host;
-  for my $from ($pms->get('EnvelopeFrom:addr',undef)) {
+  foreach my $from ($pms->get('EnvelopeFrom:addr', undef)) {
     next unless defined $from;
-
-    $from =~ tr/././s;		# bug 3366
-    if ($from =~ m/ \@ ( [^\@\s]+ \. [^\@\s]+ )/x ) {
+    $from =~ tr/.//s; # bug 3366
+    if ($from =~ m/\@([^\@\s]+\.[^\@\s]+)/) {
       $host = lc($1);
       last;
     }
@@ -650,15 +747,54 @@ sub check_dns_sender {
     return 0;
   }
 
-  dbg("dns: checking A and MX for host $host");
+  $host = idn_to_ascii($host);
+  if (exists $pms->{conf}->{uridnsbl_skip_domains}->{lc $host}) {
+    dbg("dnseval: query skipped, uridnsbl_skip_domains: $host");
+    return 0;
+  }
+  my $dom = $pms->{main}->{registryboundaries}->trim_domain($host);
+  if (exists $pms->{conf}->{uridnsbl_skip_domains}->{lc $host}) {
+    dbg("dnseval: query skipped, uridnsbl_skip_domains: $host");
+    return 0;
+  }
+  dbg("dnseval: checking A and MX for host $host");
 
-  $pms->do_dns_lookup($rule, 'A', $host);
-  $pms->do_dns_lookup($rule, 'MX', $host);
+  my $queries;
+  my $ret = $self->do_sender_lookup($pms, $rule, 'A', $host);
+  $queries++ if defined $ret;
+  $ret = $self->do_sender_lookup($pms, $rule, 'MX', $host);
+  $queries++ if defined $ret;
 
-  # cache name of host for later checking
-  $pms->{sender_host} = $host;
+  return 0 if !$queries; # no query started
+  return; # return undef for async status
+}
 
-  return 0;
+sub do_sender_lookup {
+  my ($self, $pms, $rule, $type, $host) = @_;
+
+  my $ent = {
+    rulename => $rule,
+    type => "DNSBL-Sender",
+  };
+  return $pms->{async}->bgsend_and_start_lookup(
+    $host, $type, undef, $ent, sub {
+      my ($ent, $pkt) = @_;
+      return if !$pkt; # aborted / timed out
+      $pms->rule_ready($ent->{rulename}); # mark as run, could still hit
+      foreach my $answer ($pkt->answer) {
+        next if !$answer;
+        next if $answer->type ne 'A' && $answer->type ne 'MX';
+        if ($pkt->header->rcode eq 'NXDOMAIN' ||
+            $pkt->header->rcode eq 'SERVFAIL')
+        {
+          if (++$pms->{sender_host_fail} == 2) {
+            $pms->got_hit($ent->{rulename}, "DNS: ", ruletype => "dns");
+          }
+        }
+      }
+    },
+    master_deadline => $self->{master_deadline},
+  );
 }
 
 # capability checks for "if can(Mail::SpamAssassin::Plugin::DNSEval::XXX)":
