@@ -41,7 +41,7 @@ disable something else for whatever reason... well, keep that in mind.
 
 Error messages are not unified.
 
-See <http://bugzilla.spamassassin.org/>
+See E<lt>http://bugzilla.spamassassin.org/E<gt>
 
 =head1 SEE ALSO
 
@@ -255,7 +255,7 @@ sub _validate {
     if (exists $self->{'ident-timeout'} && $self->{'ident-timeout'} <= 0) {
       die "ERROR: --ident-timeout must be > 0\n";
     }
-    ##import Net::Ident qw(ident_lookup);
+    ##Net::Ident->import(qw(ident_lookup));
   }
 
   # let's not modify %ENV here...
