@@ -21,7 +21,9 @@ if ($mode eq 'redirect') {
   open (STDOUT, ">$stdout") || die "Could not redirect STDOUT to $stdout: $!";
   open (STDERR, ">$stderr") || die "Could not redirect STDERR to $stderr: $!";
   
-  $| = 1;
+  select STDERR; $| = 1;
+  select STDOUT; $| = 1;
+
   exec { $args[0] } @args;
   die "Could not exec " . join(' ', @args) . ": $!";
 }
